import argparse

from blueness import module
from blueness.argparse.generic import sys_exit

from bluer_ugv import NAME
from bluer_ugv import env
from bluer_ugv.swallow.session.classical.ultrasonic_sensor.review import review
from bluer_ugv.swallow.session.classical.ultrasonic_sensor.testing import test
from bluer_ugv.logger import logger

NAME = module.name(__file__, NAME)

parser = argparse.ArgumentParser(NAME)
parser.add_argument(
    "task",
    type=str,
    help="review|test",
)
parser.add_argument(
    "--max_m",
    type=float,
    default=env.BLUER_UGV_ULTRASONIC_SENSOR_MAX_M,
    help="detection threshold in meters.",
)
parser.add_argument(
    "--object_name",
    type=str,
)
parser.add_argument(
    "--log",
    type=int,
    default=1,
    help="0 | 1.",
)
parser.add_argument(
    "--export",
    type=int,
    default=1,
    help="0 | 1",
)
parser.add_argument(
    "--gif",
    type=int,
    default=1,
    help="0 | 1",
)
parser.add_argument(
    "--rm_blank",
    type=int,
    default=1,
    help="0 | 1.",
)
parser.add_argument(
    "--frame_count",
    type=int,
    default=100,
    help="-1: all",
)
args = parser.parse_args()

success = False
if args.task == "review":
    success = review(
        object_name=args.object_name,
        export_gif=args.gif == 1,
        frame_count=args.frame_count,
        rm_blank=args.rm_blank == 1,
    )
elif args.task == "test":
    success = test(
        object_name=args.object_name,
        max_m=args.max_m,
        export=args.export == 1,
        export_gif=args.gif == 1,
        frame_count=args.frame_count,
        log=args.log == 1,
        rm_blank=args.rm_blank == 1,
    )
else:
    success = None

sys_exit(logger, NAME, args.task, success)



from typing import Tuple

from bluer_ugv.README.ugvs.db import dict_of_ugvs
from bluer_ugv.logger import logger


def find_server(
    hostname: str,
) -> Tuple[
    bool,  # success
    bool,  # is_server
    str,  # "0.0.0.0" if is_server else f"{server_name}.local"
]:
    for ugv_name, info in dict_of_ugvs.items():
        found: bool = False
        dict_of_computers = info.get("computers", {})
        assert isinstance(dict_of_computers, dict)

        for location, hostname_ in dict_of_computers.items():
            if hostname != hostname_:
                continue

            logger.info(f"{hostname} == {ugv_name}.{location}")
            found = True

            if location == "front":
                return True, True, "0.0.0.0"

        if not found:
            continue

        list_of_server_names = [
            hostname_
            for location, hostname_ in dict_of_computers.items()
            if location == "front"
        ]
        if not list_of_server_names:
            logger.error(f"no server (.front) onboard {ugv_name}.")
            return False, False, ""

        if len(list_of_server_names) > 1:
            logger.warning(
                "{} servers (.front) onboard {}".format(
                    len(list_of_server_names),
                    ugv_name,
                )
            )

        server_name = list_of_server_names[0]
        logger.info(f"{hostname}: server_name={server_name}")

        return True, False, f"{server_name}.local"

    logger.error(f"cannot find hostname: {hostname}.")
    return False, False, ""



from typing import List

from bluer_options.terminal import show_usage


def help_git(
    tokens: List[str],
    mono: bool,
) -> str:
    options = "designs"

    return show_usage(
        [
            "@ugv",
            "git",
            f"[{options}]",
        ],
        "@git @ugv.",
        mono=mono,
    )



from bluer_sbc.README.design import design_doc_parts

parts = {
    "470-mF": "",
    "green-terminal": "2 x",
    "LED": "green + red + yellow + 4 x blue",
    "pin-headers": "1 x (female, 2 x 40) -> 2 x 20 + 2 x (male, 1 x 40) -> 4 x 1 + 2 x 20 + 1 x (male, 2 x 40) -> 2 x 2 x 6",
    "Polyfuse": "optional",
    "pushbutton": "",
    "resistor": "7 x 330-470 Ω + 4 x 2.2 kΩ + 4 x 3.3 kΩ",
    "TVS-diode": "",
}

docs = [
    {
        "path": "../docs/swallow/digital/design/computer/shield/parts.md",
        "macros": design_doc_parts(
            dict_of_parts=parts,
            parts_reference="https://github.com/kamangir/bluer-sbc/tree/main/bluer_sbc/docs/parts",
        ),
    }
]



from typing import List, Dict, Any, Union

from bluer_ugv.README.ugvs.comparison.features.classes import (
    FeatureList,
    Feature_Comparison,
)
from bluer_ugv.README.ugvs.comparison.features.db import dict_of_feature_classes
from bluer_ugv.logger import logger


class UGV:
    def __init__(
        self,
        nickname: str,
        name: str,
        features: Dict[str, Any],
        deficiencies: List[str] = [],
        image: str = "",
        comments: List[str] = [],
    ):
        self.nickname = nickname
        self.name = name

        self.feature_list: FeatureList = FeatureList()

        for feature_name in features:
            if feature_name not in dict_of_feature_classes:
                logger.error(f"{feature_name}: feature not found.")
                assert False

        for feature_name in dict_of_feature_classes:
            if feature_name not in features:
                continue

            self.feature_list.add(
                dict_of_feature_classes[feature_name](
                    score=features[feature_name],
                )
            )

        self.deficiencies = deficiencies

        self.image = image

        self.comments = comments

    def compare(
        self,
        ugv: "UGV",
        verbose: bool = False,
    ) -> List[str]:
        similarities: List[str] = []
        differences: List[str] = []

        for feature in self.feature_list.db:
            comparison, message = feature.compare(
                ugv.feature_list.get(feature.nickname),
                self.name,
                log=verbose,
            )

            if comparison == Feature_Comparison.SIMILAR:
                similarities.append(message)
            elif comparison == Feature_Comparison.UNKNOWN:
                pass
            else:
                differences.append(
                    '<p style="color:{};">{}</p>'.format(
                        "green" if comparison == Feature_Comparison.HIGHER else "red",
                        message,
                    )
                )

        return (
            [
                '<p dir="rtl" style="text-align:right;">مشابهت: </p>',
                "<ol>",
            ]
            + [
                f'<li dir="rtl" style="text-align:right;">{line}</li>'
                for line in similarities
            ]
            + [
                "</ol>",
                '<p dir="rtl" style="text-align:right;">تفاوت‌ها:</p>',
                "<ol>",
            ]
            + [
                f'<li dir="rtl" style="text-align:right;">{line}</li>'
                for line in differences
            ]
            + [
                "</ol>",
            ]
        )

    @property
    def description(self) -> List[str]:
        return (
            [
                '<b><p dir="rtl" style="text-align:right;">{}</p></b>'.format(
                    self.name
                ),
                "<ol>",
            ]
            + [
                '<li dir="rtl" style="text-align:right;">{}</li>'.format(
                    feature.description
                )
                for feature in self.feature_list.db
            ]
            + ["</ol>"]
        )


class List_of_UGVs:
    def __init__(self):
        self.db: List[UGV] = []

    def add(
        self,
        **kw_args,
    ):
        ugv = UGV(**kw_args)
        self.db.append(ugv)

    def get(
        self,
        ugv_name: str,
    ) -> Union[UGV, None]:
        for ugv in self.db:
            if ugv_name == ugv.nickname:
                return ugv

        return None



from bluer_ugv.README.ugvs.db import dict_of_ugvs


def get(
    ugv_name: str,
    what: str,
) -> str:
    list_of_what = what.split(".")
    info = dict_of_ugvs.get(ugv_name, {})
    for what_ in list_of_what:
        info = info.get(what_, {})

    if isinstance(info, dict) and not info:
        info = ""

    if not info:
        info = "not-found"

    return str(info)



from bluer_objects.README.items import ImageItems

from bluer_ugv.README.consts import algo_docs
from bluer_ugv.README.swallow.consts import swallow_assets2
from bluer_ugv.README.swallow.digital.algo import driving, navigation, tracking, yolo

docs = (
    [
        {
            "path": "../docs/swallow/digital/algo",
        },
    ]
    + driving.docs
    + navigation.docs
    + tracking.docs
    + yolo.docs
)



from bluer_ugv.README.swallow.digital import algo, design

docs = (
    [
        {
            "path": "../docs/swallow/digital",
        },
        {
            "path": "../docs/swallow/digital/manual.md",
        },
    ]
    + design.docs
    + algo.docs
)



from bluer_objects.README.items import ImageItems
from bluer_objects.README.consts import assets
from bluer_objects.README.consts import assets_url

from bluer_ugv.README.consts import algo_docs
from bluer_ugv.README.swallow.consts import swallow_assets2

tracker_assets2 = assets_url(
    "tracker",
    volume=2,
)

docs = [
    {
        "path": "../docs/swallow/digital/algo/tracking",
    },
    {
        "path": "../docs/swallow/digital/algo/tracking/validations",
    },
    {
        "path": "../docs/swallow/digital/algo/tracking/validations/one.md",
        "items": ImageItems(
            {
                f"{swallow_assets2}/target-selection.png": f"{algo_docs}/socket.md",
            }
        ),
    },
    {
        "path": "../docs/swallow/digital/algo/tracking/validations/two.md",
        "items": ImageItems(
            {
                f"{tracker_assets2}/validations/two/socket-3.png": f"{algo_docs}/socket.md",
                "{}/{}/{}.gif".format(
                    assets,
                    "swallow-debug-2025-12-11-17-23-59-6bhm5n",
                    "swallow-debug-2025-12-11-17-23-59-6bhm5n",
                ): "",
            }
        ),
    },
    {
        "path": "../docs/swallow/digital/algo/tracking/validations/three.md",
        "items": ImageItems(
            {
                "{}/{}/{}.gif".format(
                    assets,
                    "swallow-debug-2025-12-12-14-46-34-m8ahlp",
                    "swallow-debug-2025-12-12-14-46-34-m8ahlp",
                ): "",
            }
        ),
    },
]



from bluer_objects.README.items import ImageItems, Items

from bluer_ugv.README.ugvs.db import dict_of_ugvs
from bluer_ugv.README.validations.db import dict_of_validations

docs = [
    {
        "path": f"../docs/UGVs/{ugv_name}.md",
        "cols": info.get("cols", 3),
        "items": ImageItems(
            {
                item: ""
                for item in info.get(
                    "items",
                    [],
                )
            }
        ),
        "macros": {
            "class:::": "[{class_name}](../{class_name}){info}".format(
                class_name=(
                    info["class"][0]
                    if isinstance(info["class"], list)
                    else info["class"]
                ),
                info=(
                    " ".join([""] + info["class"][1:])
                    if isinstance(info["class"], list)
                    else ""
                ),
            ),
            "computers:::": (
                (
                    [
                        "- computer(s):",
                    ]
                    + [
                        "    - {}: `{}`".format(
                            location,
                            computer_name,
                        )
                        for location, computer_name in info["computers"].items()
                    ]
                )
                if "computers" in info
                else []
            ),
            "comments:::": [
                f"- {comment}"
                for comment in info.get(
                    "comments",
                    [],
                )
            ],
            "tagline:::": info.get("tagline", ""),
            "validations:::": [
                (lambda thing: f"validations: {thing}" if thing else "")(
                    ", ".join(
                        sorted(
                            [
                                f"[`{validation_name}`](../validations/{validation_name}.md)"
                                for validation_name, info in dict_of_validations.items()
                                if any(
                                    ugv_name_.startswith(f"{ugv_name}:")
                                    for ugv_name_ in info["ugv_name"]
                                )
                            ]
                        )
                    )
                ),
            ],
        },
    }
    for ugv_name, info in dict_of_ugvs.items()
] + [
    {
        "path": "../docs/UGVs",
        "items": Items(
            [
                {
                    "name": ugv_name,
                    "marquee": (
                        (
                            lambda url: (
                                url if url.endswith("?raw=true") else f"{url}/?raw=true"
                            )
                        )(info["items"][-1])
                        if info["items"]
                        else ""
                    ),
                    "url": f"./{ugv_name}.md",
                }
                for ugv_name, info in dict_of_ugvs.items()
            ]
        ),
        "macros": {
            "list:::": [
                f"- [{ugv_name}](./{ugv_name}.md)"
                for ugv_name in sorted(
                    dict_of_ugvs.keys(),
                    key=lambda ugv_name: dict_of_ugvs[ugv_name]["order"],
                )
            ],
        },
    }
]



from typing import List, Optional
import subprocess
import shlex
import time

from bluer_options.logger import crash_report
from bluer_options.logger.config import log_list
from bluer_objects import file

from bluer_ugv.swallow.session.classical.screen.video.engine import VideoEngine
from bluer_ugv.logger import logger


class VideoPlayer:
    def __init__(
        self,
        dryrun: bool = False,
        engine: VideoEngine = VideoEngine.VLC,
    ):
        self.process: Optional[subprocess.Popen] = None
        self.current_file: Optional[str] = None

        self.paused = False
        self.dryrun = dryrun
        assert isinstance(engine, VideoEngine)
        self.engine: VideoEngine = engine

        logger.info(
            "{} created on {}{}.".format(
                self.__class__.__name__,
                self.engine.name.lower(),
                " [dryrun]" if dryrun else "",
            )
        )

    def pause(self) -> bool:
        if not self.dryrun:
            if self.process and self.process.poll() is None:
                if not self.engine.pause(self.process):
                    return False

        logger.info(
            "{}.{}".format(
                self.__class__.__name__,
                "resume" if self.paused else "pause",
            )
        )
        self.paused = not self.paused

        return True

    def play(
        self,
        filename: str,
        loop: bool = False,
        audio: bool = False,
        fullscreen: bool = True,
        verbose: bool = True,
    ) -> bool:
        if not file.exists(filename):
            logger.error(f"file not found: {filename}")
            return False

        self.stop()

        command = self.engine.play_command(
            filename=filename,
            fullscreen=fullscreen,
            loop=loop,
            audio=audio,
        )
        logger.info(f"running on {self.engine.name.lower()}: {command}")

        if not self.dryrun:
            try:
                # pylint: disable=consider-using-with
                self.process = subprocess.Popen(
                    shlex.split(command),
                    stdin=subprocess.PIPE if self.engine == VideoEngine.MPV else None,
                    stdout=None if verbose else subprocess.DEVNULL,
                    stderr=None if verbose else subprocess.DEVNULL,
                )

                logger.info(
                    f"pid={self.process.pid}, "
                    f"stdin={self.process.stdin}, returncode={self.process.returncode}"
                )

            except Exception as e:
                crash_report(f"failed to run: {e}")
                self.process = None
                return False

            if not self.process:
                logger.error("process is None.")
                return False

        self.current_file = filename

        logger.info(
            "{}.play({}{})".format(
                self.__class__.__name__,
                "loop: " if loop else "",
                filename,
            )
        )

        return True

    def stop(self) -> bool:
        if not self.dryrun and self.process and self.process.poll() is None:
            if not self.engine.stop(self.process):
                return False

        self.process = None
        logger.info(f"{self.__class__.__name__}.stop")
        return True



dict_of_parts = {
    "4-ch-transceiver": "",
    "2xAA-battery-holder": "",
    "4xAA-battery-holder": "",
    "yellow-gearbox-dc-motor": "4 x 270 RPM",
    "yellow-wheels": "x 4",
    "L-1x2": "x 8",
    "nuts-bolts-spacers": "M3: ({})".format(
        " + ".join(
            [
                "10 x 6 mm bolt",
                "1 x 5 mm bolt",
                "8 x 30 mm bolt",
                "20 x nut",
                "1 x 5 mm spacer",
                "2 x 10 mm spacer",
            ]
        )
    ),
    "ni-mh-battery": "x 4",
    "plexiglass": "14 cm x 9.5 cm",
    "PCB-single-14x9_5": "",
    "mt-3608": "",
    "arduino-nano": "",
    "tb6612": "",
    "small-dc-switch": "",
    "pin-headers": "2 x (male, 1 x 40) -> 2 x 4 + 2 x 6 + 2 x 15",
    "double-sided-tape": "",
    "electrical-tape": "",
    "micro-usb-cable": "",
}



from bluer_objects.README.items import ImageItems

from bluer_ugv.README.rangin.consts import rangin_assets2

docs = [
    {
        "path": "../docs/rangin/specs.md",
        "items": ImageItems(
            {
                f"{rangin_assets2}/20251116_124842.jpg": "",
            }
        ),
    },
]



from bluer_objects.README.alias import list_of_aliases

from bluer_ugv import NAME
from bluer_ugv.README.items import items
from bluer_ugv.README.shortcuts import items as shortcuts_items


docs = [
    {
        "path": "../..",
        "items": items,
        "macros": {
            "shortcuts:::": shortcuts_items,
            "aliases:::": list_of_aliases(NAME),
        },
    },
    {
        "path": "../docs",
    },
]



from bluer_ugv.README.arzhang import docs as arzhang_docs
from bluer_ugv.README.eagle import docs as eagle_docs
from bluer_ugv.README.fire import docs as fire_docs
from bluer_ugv.README.rangin import docs as rangin_docs
from bluer_ugv.README.ravin import docs as ravin_docs
from bluer_ugv.README import (
    aliases,
    beast,
    root,
    releases,
    swallow,
)
from bluer_ugv.README.computer import docs as computer
from bluer_ugv.README.ugvs import docs as ugvs
from bluer_ugv.README.validations import docs as validations

docs = (
    root.docs
    + aliases.docs
    + arzhang_docs.docs
    + beast.docs
    + eagle_docs.docs
    + fire_docs.docs
    + rangin_docs.docs
    + ravin_docs.docs
    + releases.docs
    + computer.docs
    + ugvs.docs
    + swallow.docs
    + validations.docs
)



from bluer_ugv.README.ravin.ravin3 import docs
from bluer_ugv.README.ravin.items import items
from bluer_ugv.README.ravin.ravin4 import docs

docs = (
    [
        {
            "items": items,
            "path": "../docs/ravin",
        },
    ]
    + docs.docs
    + docs.docs
)



from bluer_ugv.README.ugvs.comparison.features.control import UGV_Control
from bluer_ugv.README.ugvs.comparison.features.cost import UGV_Cost
from bluer_ugv.README.ugvs.comparison.features.size import UGV_Size
from bluer_ugv.README.ugvs.comparison.ugvs.classes import List_of_UGVs

list_of_ugvs = List_of_UGVs()

list_of_ugvs.add(
    nickname="arzhang",
    name="محصول ما",
    features={
        "concealment": True,
        "control": UGV_Control.AI,
        "cost": UGV_Cost.LOW,
        "payload": 40,
        "range": 10,
        "ps": True,
        "sanction_proof": True,
        "size": UGV_Size.SMALL,
        "speed": 4,
        "swarm": True,
        "uv_delivery": True,
    },
)

list_of_ugvs.add(
    nickname="nazir",
    image="nazir.jpg",
    name="ربات موشک‌انداز نذیر",
    features={
        "control": UGV_Control.RC,
        "cost": UGV_Cost.MEDIUM,
        "payload": 700,
        "range": 4,
        "sanction_proof": True,
        "size": UGV_Size.MEDIUM,
    },
)

list_of_ugvs.add(
    nickname="heydar",
    name="ربات حیدر",
    image="heydar.jpg",
    features={
        "concealment": True,
        "control": UGV_Control.AI,
        "cost": UGV_Cost.LOW,
        "payload": 40,
        "range": 10,
        "sanction_proof": True,
        "size": UGV_Size.SMALL,
        "speed": 60,
        "swarm": True,
    },
    deficiencies=[
        "انتقال قدرت: زنجیر",
    ],
    comments=[
        "بیشترین تشابه",
    ],
)


list_of_ugvs.add(
    nickname="karakal",
    name="ربات جنگجوی هوشمند کاراکال",
    image="karakal.jpg",
    features={
        "control": UGV_Control.AI,
        "cost": UGV_Cost.MEDIUM,
        "sanction_proof": True,
        "size": UGV_Size.MEDIUM,
        "speed": 30,
        "range": 0.5,
    },
)

list_of_ugvs.add(
    nickname="qasem",
    name="ربات قاسم",
    image="qasem.jpg",
    features={
        "concealment": True,
        "cost": UGV_Cost.MEDIUM,
        "sanction_proof": True,
        "size": UGV_Size.MEDIUM,
        "uv_delivery": True,
    },
    deficiencies=[],
)

list_of_ugvs.add(
    nickname="arya",
    name="ربات آریا",
    image="arya.jpg",
    features={
        "control": UGV_Control.AI,
        "cost": UGV_Cost.MEDIUM,
        "sanction_proof": True,
        "size": UGV_Size.MEDIUM,
        "speed": 50,
        "uv_delivery": True,
    },
    deficiencies=[],
)

list_of_ugvs.add(
    nickname="sepah",
    name="ربات جنگ میدانی سپاه",
    features={
        "sanction_proof": True,
    },
    deficiencies=[],
)


list_of_ugvs.add(
    nickname="raad1",
    name="متلاشی‌کننده بمب و تله انفجاری رعد ۱",
    image="raad1.jpg",
    features={
        "control": UGV_Control.RC,
        "cost": UGV_Cost.MEDIUM,
        "sanction_proof": True,
        "size": UGV_Size.SMALL,
    },
    deficiencies=[],
)

list_of_ugvs.add(
    nickname="uran_6",
    name="Uran-6",
    image="Uran-6.jpeg",
    features={
        "control": UGV_Control.AI,
        "cost": UGV_Cost.HIGH,
        "range": 1,
        "size": UGV_Size.LARGE,
        "speed": 5,
    },
    deficiencies=[],
)

list_of_ugvs.add(
    nickname="uran_9",
    name="Uran-9",
    image="Uran-9.jpg",
    features={
        "control": UGV_Control.AI,
        "cost": UGV_Cost.HIGH,
        "range": 1,
        "size": UGV_Size.LARGE,
        "speed": 133,
    },
    deficiencies=[],
)

list_of_ugvs.add(
    nickname="uran_14",
    name="Uran-14",
    image="Uran-14.jpg",
    features={
        "control": UGV_Control.AI,
        "cost": UGV_Cost.HIGH,
        "payload": 2700,
        "range": 1.5,
        "size": UGV_Size.LARGE,
        "speed": 10,
    },
    deficiencies=[],
)

list_of_ugvs.add(
    nickname="themis_9",
    name="THeMIS-9",
    image="THeMIS-9.jpeg",
    features={
        "control": UGV_Control.AI,
        "cost": UGV_Cost.HIGH,
        "payload": 1630,
        "range": 1.5,
        "size": UGV_Size.LARGE,
        "speed": 20,
    },
    deficiencies=[],
)

list_of_ugvs.add(
    nickname="type_x",
    name="Type-X",
    image="Type-X.jpg",
    features={
        "control": UGV_Control.AI,
        "cost": UGV_Cost.HIGH,
        "payload": 4100,
        "size": UGV_Size.LARGE,
        "speed": 80,
    },
    deficiencies=[],
)


list_of_ugvs.add(
    nickname="centaur",
    name="Centaur",
    image="Centaur.jpg",
    features={
        "control": UGV_Control.RC,
        "cost": UGV_Cost.HIGH,
        "payload": 14.5,
        "range": 0.8,
        "size": UGV_Size.MEDIUM,
        "speed": 4,
    },
    deficiencies=[],
)


list_of_ugvs.add(
    nickname="xm1219",
    name="XM1219",
    image="XM1219.jpeg",
    features={
        "control": UGV_Control.RC,
        "cost": UGV_Cost.HIGH,
        "size": UGV_Size.LARGE,
    },
    deficiencies=[],
)

list_of_ugvs.add(
    nickname="talon",
    name="Foster-Miller TALON",
    image="talon.jpg",
    features={
        "control": UGV_Control.RC,
        "cost": UGV_Cost.HIGH,
        "range": 1.2,
        "size": UGV_Size.MEDIUM,
    },
    deficiencies=[],
)

list_of_ugvs.add(
    nickname="gladiator",
    name="Gladiator TUGV",
    image="Gladiator.jpg",
    features={
        "control": UGV_Control.RC,
        "cost": UGV_Cost.HIGH,
        "size": UGV_Size.LARGE,
    },
    deficiencies=[],
)

list_of_ugvs.add(
    nickname="ukap",
    name="UKAP",
    image="UKAP.jpg",
    features={
        "control": UGV_Control.RC,
        "cost": UGV_Cost.HIGH,
        "size": UGV_Size.LARGE,
    },
    deficiencies=[],
)

list_of_ugvs.add(
    nickname="ripsaw",
    name="Ripsaw",
    image="Ripsaw.jpg",
    features={
        "control": UGV_Control.AI,
        "cost": UGV_Cost.HIGH,
        "payload": 910,
        "size": UGV_Size.LARGE,
        "speed": 105,
    },
    deficiencies=[],
)

list_of_ugvs.add(
    nickname="teodor",
    name="tEODor",
    image="tEODor.png",
    features={
        "concealment": True,
        "control": UGV_Control.RC,
        "cost": UGV_Cost.HIGH,
        "payload": 100,
        "size": UGV_Size.MEDIUM,
        "uv_delivery": True,
    },
    deficiencies=[],
)


list_of_ugvs.add(
    nickname="black_knight",
    name="Black Knight",
    image="black_knight.jpeg",
    features={
        "control": UGV_Control.AI,
        "cost": UGV_Cost.HIGH,
        "size": UGV_Size.LARGE,
        "speed": 77,
    },
    deficiencies=[],
)

list_of_ugvs.add(
    nickname="guardium",
    name="Guardium",
    image="Guardium.jpg",
    features={
        "control": UGV_Control.AI,
        "cost": UGV_Cost.HIGH,
        "size": UGV_Size.LARGE,
    },
    deficiencies=[],
)

list_of_ugvs.add(
    nickname="milica",
    name="Milos & Milica",
    image="Milos.jpg",
    features={
        "control": UGV_Control.RC,
        "cost": UGV_Cost.MEDIUM,
        "payload": 100,
        "range": 3,
        "size": UGV_Size.LARGE,
        "speed": 12.5,
    },
    deficiencies=[],
)

list_of_ugvs.add(
    nickname="taifun_m",
    name="Taifun-M",
    image="Taifun-M.jpg",
    features={
        "control": UGV_Control.RC,
        "cost": UGV_Cost.HIGH,
        "size": UGV_Size.LARGE,
        "uv_delivery": True,
    },
    deficiencies=[],
)


list_of_ugvs.add(
    nickname="jackal",
    name="Jackal",
    image="Jackal.jpeg",
    features={
        "control": UGV_Control.AI,
        "cost": UGV_Cost.HIGH,
        "payload": 20,
        "size": UGV_Size.MEDIUM,
        "speed": 7.2,
    },
    deficiencies=[],
)

list_of_ugvs.add(
    nickname="avantguard",
    name="AvantGuard UGCV",
    image="avantguard.jpg",
    features={
        "control": UGV_Control.AI,
        "cost": UGV_Cost.HIGH,
        "size": UGV_Size.MEDIUM,
        "speed": 20,
    },
    deficiencies=[],
)


list_of_ugvs.add(
    nickname="sr-lambda",
    name="SIM-RACAR-Lambda",
    image="SIM-RACAR-Lambda.jpg",
    features={
        "control": UGV_Control.AI,
        "cost": UGV_Cost.HIGH,
        "payload": 45,
        "range": 1,
        "size": UGV_Size.LARGE,
    },
    deficiencies=[],
)

list_of_ugvs.add(
    nickname="sharp_claw_1",
    name="Sharp Claw 1",
    image="Sharp-Claw-1.jpg",
    features={
        "control": UGV_Control.RC,
        "cost": UGV_Cost.HIGH,
        "range": 6,
        "size": UGV_Size.LARGE,
        "speed": 9,
        "uv_delivery": False,
    },
    deficiencies=[],
)

list_of_ugvs.add(
    nickname="sharp_claw_2",
    name="Sharp Claw 2",
    image="Sharp-Claw-2.jpg",
    features={
        "control": UGV_Control.RC,
        "cost": UGV_Cost.HIGH,
        "payload": 120,
        "range": 50,
        "size": UGV_Size.LARGE,
        "speed": 30,
        "uv_delivery": True,
    },
    deficiencies=[],
)

list_of_ugvs.add(
    nickname="robattle",
    name="Robattle",
    image="Robattle.jpeg",
    features={
        "control": UGV_Control.AI,
        "cost": UGV_Cost.HIGH,
        "payload": 3000,
        "size": UGV_Size.LARGE,
    },
    deficiencies=[],
)

list_of_ugvs.add(
    nickname="template",
    name="template",
    features={
        "control": ...,
        "cost": ...,
        "payload": ...,
        "range": ...,
        "size": ...,
        "speed": ...,
        "swarm": ...,
        "uv_delivery": ...,
    },
    deficiencies=[],
)



from bluer_objects.README.items import ImageItems

from bluer_ugv.README.swallow.consts import swallow_assets2
from bluer_ugv.README.computer import box, power, schematics, testing
from bluer_ugv.README.computer.shield import docs as shield


docs = (
    [
        {
            "path": "../docs/swallow/digital/design/computer",
            "items": ImageItems(
                {
                    f"{swallow_assets2}/20251121_124335.jpg": "",
                }
            ),
        },
        {
            "path": "../docs/swallow/digital/design/computer/naming.md",
        },
        {
            "path": "../docs/swallow/digital/design/computer/terraform.md",
        },
        {
            "path": "../docs/swallow/digital/design/computer/connectors-v1.md",
        },
    ]
    + box.docs
    + shield.docs
    + power.docs
    + schematics.docs
    + testing.docs
)



from blueness import module
import numpy as np

from bluer_options.terminal import hr
from bluer_algo.socket.connection import SocketConnection
from bluer_algo.tracker.classes.target import Target

from bluer_ugv import NAME
from bluer_ugv.logger import logger


NAME = module.name(__file__, NAME)

DEFAULT_TARGETING_PORT = 8002


def select_target(
    host: str,
    loop: bool = True,
    port: int = DEFAULT_TARGETING_PORT,
) -> bool:
    logger.info(
        "{}.select_target on {} port={}{}".format(
            NAME,
            host,
            port,
            " on a loop." if loop else "",
        )
    )

    try:
        while loop:
            socket = SocketConnection.listen_on(port=port)
            success, image = socket.receive_data(np.ndarray)
            if not success:
                return success

            success, track_window = Target.select(
                image,
                title=f"select the target for {host} port={port}...",
            )
            if not success:
                return success

            socket = SocketConnection.connect_to(
                host,
                port=port,
            )
            if not socket.send_data(track_window):
                return False

            hr_line = hr(
                width=12,
                mono=True,
            )
            logger.info(f"{hr_line} Ctrl+C to exit {hr_line}")
    except KeyboardInterrupt:
        logger.info("Ctrl+C, stopping.")

    return True



from bluer_ugv.README.consts import assets_url
from bluer_ugv.README.swallow.consts import swallow_assets2
from bluer_ugv.README.arzhang.consts import (
    arzhang_assets,
    arzhang_assets2,
    arzhang2_assets2,
)
from bluer_ugv.README.rangin.consts import rangin_assets2, rangin_mechanical_design

dict_of_ugvs = {
    "swallow": {
        "order": 1,
        "cols": 1,
        "items": [
            f"{swallow_assets2}/20250701_2206342_1.gif",
            f"{swallow_assets2}/20250913_203635~2_1.gif",
            f"{swallow_assets2}/20251216_205015.jpg",
        ],
        "tagline": "the first one.",
        "class": "swallow",
        "computers": {
            "front": "swallow2",  #  `swallow` was used for Ubuntu experiments
        },
    },
    "arzhang": {
        "order": 2,
        "items": [
            f"{arzhang_assets2}/20251209_111322.jpg",
        ],
        "tagline": "the first [arzhang](../arzhang).",
        "class": "arzhang",
        "computers": {
            "front": "sparrow",
        },
        "comments": [
            "disassembled 🛑",
        ],
    },
    "arzhang2": {
        "order": 3,
        "items": [
            f"{arzhang_assets2}/20251210_154513.jpg",
            f"{arzhang2_assets2}/20260108_175635.jpg",
            f"{arzhang2_assets2}/20260109_133216.jpg",
            f"{arzhang_assets2}/IMG_0112.JPG",
            f"{arzhang_assets2}/IMG_0118.JPG",
        ],
        "tagline": "an [arzhang](../arzhang) with updated body design.",
        "class": "arzhang",
        "computers": {
            "front": "sparrow2",
            "back": "sparrow3-back",
        },
    },
    "arzhang3": {
        "order": 4,
        "items": sorted(
            [
                f"{arzhang_assets}/20251107_175506.jpg",
                f"{arzhang_assets2}/20251128_175614.jpg",
                f"{arzhang_assets2}/20251202_100317.jpg",
                f"{arzhang_assets2}/20251202_101031.jpg",
                f"{arzhang_assets2}/20251128_113314.jpg",
                f"{arzhang_assets2}/20251128_151952.jpg",
                f"{arzhang_assets2}/20251128_155616.jpg",
                f"{arzhang_assets2}/20251130_140103.jpg",
                f"{arzhang_assets2}/20251203_112602.jpg",
                f"{arzhang_assets2}/20251210_154654.jpg",
            ]
        ),
        "tagline": "the first [arzhang](../arzhang) with two computers.",
        "class": "arzhang",
        "computers": {
            "front": "arzhang3-front",
            "back": "sparrow3-back",
        },
        "comments": [
            "-> [rangin](./rangin.md) 🛑",
        ],
    },
    "rangin": {
        "order": 5,
        "items": [
            f"{rangin_mechanical_design}/robot.png",
            f"{rangin_assets2}/20251221_125120.jpg",
            f"{rangin_assets2}/20251222_104404.jpg",
            f"{rangin_assets2}/20251222_110634.jpg",
            f"{rangin_assets2}/20251222_160737.jpg",
            f"{rangin_assets2}/20251222_165829.jpg",
            f"{rangin_assets2}/20251222_173509.jpg",
            f"{rangin_assets2}/IMG-20251223-WA0001.jpg",
            f"{rangin_assets2}/IMG-20251223-WA0009.jpg",
            f"{rangin_assets2}/IMG-20251223-WA0010.jpg",
            f"{rangin_assets2}/20251224_180302.jpg",
            f"{rangin_assets2}/20251224_180308.jpg",
            f"{rangin_assets2}/20251224_180319.jpg",
            f"{rangin_assets2}/20251224_182306~3_1.gif",
            f"{rangin_assets2}/2025-12-25-lab.png",
            f"{rangin_assets2}/20251226_184718.jpg",
            f"{rangin_assets2}/20251226_184008.jpg",
            f"{rangin_assets2}/20251228_125134.jpg",
            f"{rangin_assets2}/20251231_111615.jpg",
            f"{rangin_assets2}/20251231_111656.jpg",
            f"{rangin_assets2}/20251231_111815.jpg",
            f"{rangin_assets2}/20251231_111923.jpg",
            f"{rangin_assets2}/20251231_112447.jpg",
            f"{rangin_assets2}/2026-01-26-d1.jpg",
            f"{rangin_assets2}/2026-01-26-d2.jpg",
            f"{rangin_assets2}/2026-01-26-d3.jpg",
            f"{rangin_assets2}/Screenshot-2026-02-03-11-15-43.png",
            f"{rangin_assets2}/Screenshot-2026-02-03-19-46-15.png",
            f"{rangin_assets2}/IMG_0071.JPG",
            f"{rangin_assets2}/IMG_0034.JPG",
            f"{rangin_assets2}/IMG_0033.JPG",
        ],
        "tagline": "an ad robot.",
        "class": "rangin",
        "computers": {
            "front": "arzhang3-front",
            "top": "rangin-top2",
        },
        "comments": [
            "built on [arzhang3](./arzhang3.md).",
        ],
    },
    "arzhang4": {
        "order": 6,
        "items": [
            assets_url(
                suffix="bluer-sbc/parts/GM6558/01.jpg",
                volume=2,
            ),
        ],
        "tagline": "the first [arzhang](../arzhang) with [external motors](https://github.com/kamangir/bluer-sbc/blob/main/bluer_sbc/docs/parts/DC-gearboxed-motor-12V-120RPM.md).",
        "class": [
            "arzhang",
            "with external motors.",
        ],
        "comments": [
            "build is ongoing. 🔥",
        ],
    },
}



from typing import List
from tqdm import tqdm, trange

from blueness import module
from bluer_options.logger import log_list
from bluer_objects import storage, objects
from bluer_objects.metadata import get_from_object
from bluer_objects.storage.policies import DownloadPolicy
from bluer_algo.image_classifier.dataset.dataset import ImageClassifierDataset

from bluer_ugv import NAME
from bluer_ugv import env
from bluer_ugv.logger import logger

NAME = module.name(__file__, NAME)


def combine(
    object_name: str,
    count: int = -1,
    download: bool = True,
    log: bool = True,
    verbose: bool = False,
    recent: bool = True,
    sequence: int = -1,
    split: bool = True,
    test_ratio: float = 0.1,
    train_ratio: float = 0.8,
    explicit_dataset_object_names: str = "not-given",
) -> bool:
    eval_ratio = 1 - train_ratio - test_ratio
    if eval_ratio <= 0:
        logger.error(f"eval_ratio = {eval_ratio:.2f} <= 0")
        return False

    logger.info(
        "{}.combine({}{}{}{}) -{}{}> {}".format(
            NAME,
            "all" if count == -1 else f"count={count}",
            ",download" if download else "",
            ",recent" if recent else "",
            ",split" if recent else "",
            (
                "train={:.2f}/eval={:.2f}/test={:.2f}-".format(
                    train_ratio,
                    eval_ratio,
                    test_ratio,
                )
                if split
                else ""
            ),
            ("" if sequence == -1 else f"sequence={sequence}-"),
            object_name,
        )
    )

    if explicit_dataset_object_names != "not-given":
        list_of_dataset_object_names = explicit_dataset_object_names.split(",")
    else:
        logger.info(
            "reading from  {} ...".format(env.BLUER_UGV_SWALLOW_NAVIGATION_DATASET_LIST)
        )
        list_of_dataset_object_names: List[str] = get_from_object(
            object_name=env.BLUER_UGV_SWALLOW_NAVIGATION_DATASET_LIST,
            key="dataset-list",
            default=[],
            download=download,
        )

    if count != -1:
        if recent:
            list_of_dataset_object_names = list_of_dataset_object_names[-count:]
        else:
            list_of_dataset_object_names = list_of_dataset_object_names[:count]

    log_list(
        logger,
        "combining",
        list_of_dataset_object_names,
        "dataset(s)",
        itemize=True,
    )

    if download:
        for dataset_object_name in tqdm(list_of_dataset_object_names):
            logger.info(f"downloading {dataset_object_name} ...")
            if not storage.download(
                dataset_object_name,
                policy=DownloadPolicy.DOESNT_EXIST,
                log=verbose,
            ):
                return False

    success, list_of_datasets = ImageClassifierDataset.load_list(
        list_of_dataset_object_names,
        log=log,
    )
    if not success:
        return success

    if sequence != -1:
        for index in trange(len(list_of_datasets)):
            success, list_of_datasets[index] = list_of_datasets[index].sequence(
                length=sequence,
                object_name=objects.unique_object(
                    "{}-{}X".format(
                        list_of_datasets[index].object_name,
                        sequence,
                    ),
                ),
                log=log,
                verbose=verbose,
            )
            if not success:
                return success

    success, dataset = ImageClassifierDataset.combine(
        list_of_datasets,
        object_name=object_name,
        split=split,
        test_ratio=test_ratio,
        train_ratio=train_ratio,
    )
    if not success:
        return success

    return dataset.save(
        metadata={
            "contains": list_of_dataset_object_names,
        },
        log=True,
    )



import argparse

from blueness import module
from blueness.argparse.generic import sys_exit

from bluer_ugv import NAME
from bluer_ugv.README.ugvs.get import get
from bluer_ugv.logger import logger

NAME = module.name(__file__, NAME)

parser = argparse.ArgumentParser(NAME)
parser.add_argument(
    "task",
    type=str,
    help="get",
)
parser.add_argument(
    "--ugv_name",
    type=str,
)
parser.add_argument(
    "--what",
    type=str,
)
args = parser.parse_args()

success = False
if args.task == "get":
    success = True
    print(
        get(
            ugv_name=args.ugv_name,
            what=args.what,
        )
    )
else:
    success = None

sys_exit(logger, NAME, args.task, success)



import shlex
from enum import Enum, auto
import subprocess
import time
import socket

from bluer_objects.graphics.screen import get_size
from bluer_ugv.logger import logger


class VideoEngine(Enum):
    MPV = auto()
    VLC = auto()

    def pause(self, process: subprocess.Popen):
        # MPV: use stdin
        if self == VideoEngine.MPV:
            try:
                if not process.stdin:
                    logger.error("mpv pause failed: no stdin")
                    return False
                process.stdin.write(b"p")
                process.stdin.flush()
            except Exception as e:
                logger.error(f"mpv pause exception: {e}")
                return False
            return True

        # VLC: use RC TCP
        if self == VideoEngine.VLC:
            try:
                s = socket.create_connection(("127.0.0.1", 41940), 0.5)
                s.sendall(b"pause\n")
                s.close()
                return True
            except Exception as e:
                logger.error(f"vlc pause failed: {e}")
                return False

        logger.error(f"{self}: unknown video engine.")
        return False

    def play_command(
        self,
        filename: str,
        fullscreen: bool = True,
        loop: bool = False,
        audio: bool = False,
    ) -> str:
        screen_height, screen_width = get_size()
        logger.info("screen size: {}x{}".format(screen_height, screen_width))

        if self == VideoEngine.MPV:
            logger.info('press "q" to quit mpv.')

            return " ".join(
                [
                    "mpv",
                    "--no-border",
                    "--background=black",  # FIXED
                    "--keepaspect=yes",
                    "--no-keepaspect-window",
                    "--geometry=0:0",
                    (f"--autofit={screen_width}x{screen_height}" if fullscreen else ""),
                    "--loop" if loop else "",
                    "--no-audio" if not audio else "",
                    shlex.quote(filename),
                ]
            )

        if self == VideoEngine.VLC:
            logger.info('press "Enter" to quit vlc.')

            return " ".join(
                [
                    "sudo -u pi",
                    "cvlc",
                    "--fullscreen",
                    "--no-video-title-show",
                    "--no-audio-capture",
                    "--video-on-top",
                    "--no-osd",
                    "--loop" if loop else "",
                    "--no-audio" if not audio else "",
                    "--extraintf",
                    "rc",
                    "--rc-host=127.0.0.1:41940",
                    shlex.quote(filename),
                ]
            )

        return "this-should-not-happen"

    def stop(
        self,
        process: subprocess.Popen,
    ) -> bool:
        if self == VideoEngine.MPV:
            try:
                if process.stdin:
                    process.stdin.write(b"q")
                    process.stdin.flush()
            except Exception as e:
                logger.error(f"mpv quit failed: {e}")
                return False

        if self == VideoEngine.VLC:
            try:
                s = socket.create_connection(("127.0.0.1", 41940), 0.5)
                s.sendall(b"quit\n")
                s.close()
                logger.info("vlc: sent 'quit' via TCP RC.")
            except Exception as e:
                logger.error(f"vlc rc quit failed: {e}")
                return False

        time.sleep(0.3)

        try:
            if process.poll() is None:
                process.kill()
        except Exception as e:
            logger.warning(f"process.kill failed: {e}")
            return False

        return True



from bluer_sbc.parts.db import db_of_parts
from bluer_sbc.parts.consts import parts_url_prefix

from bluer_ugv.designs.swallow.parts import dict_of_parts

docs = [
    {
        "path": "../docs/swallow/digital/design/parts.md",
        "items": db_of_parts.as_images(
            dict_of_parts,
            reference=parts_url_prefix,
        ),
        "macros": {
            "parts:::": db_of_parts.as_list(
                dict_of_parts,
                reference=parts_url_prefix,
                log=False,
            ),
        },
    },
]



from typing import Tuple
import numpy as np

from bluer_options.timer import Timer
from bluer_algo.env import BLUER_ALGO_TRACKER_DEFAULT_ALGO
from bluer_algo.tracker.classes.target import Target
from bluer_algo.tracker.factory import get_tracker_class
from bluer_algo.socket.message import SocketMessage
from bluer_sbc.imager.camera import instance as camera

from bluer_ugv import env
from bluer_ugv.swallow.session.classical.camera.generic import ClassicalCamera
from bluer_ugv.swallow.session.classical.config import ClassicalConfig
from bluer_ugv.swallow.session.classical.keyboard.classes import ClassicalKeyboard
from bluer_ugv.swallow.session.classical.leds import ClassicalLeds
from bluer_ugv.swallow.session.classical.setpoint.classes import ClassicalSetPoint
from bluer_ugv.swallow.session.classical.mode import OperationMode
from bluer_ugv.swallow.targeting import DEFAULT_TARGETING_PORT
from bluer_ugv.logger import logger


class ClassicalTrackingCamera(ClassicalCamera):
    def __init__(
        self,
        config: ClassicalConfig,
        keyboard: ClassicalKeyboard,
        leds: ClassicalLeds,
        setpoint: ClassicalSetPoint,
        object_name: str,
    ):
        super().__init__(config, keyboard, leds, setpoint, object_name)

        self.track_window: Tuple[int, int, int, int] = None

        self.tracking_timer = Timer(
            period=env.BLUER_UGV_CAMERA_ACTION_PERIOD,
            name="{}.tracking".format(self.__class__.__name__),
            log=True,
        )

        self.tracker = get_tracker_class(BLUER_ALGO_TRACKER_DEFAULT_ALGO)[1]()

    def initialize(self) -> bool:
        if not super().initialize():
            return False

        return self.select_target()

    def select_target(self) -> bool:
        success, image = camera.capture(
            close_after=False,
            open_before=False,
            log=True,
        )
        if not success:
            return success

        self.leds.set_all(True)
        success, self.track_window = Target.select(
            np.flip(image, axis=2),
            local=False,
            port=DEFAULT_TARGETING_PORT,
        )
        self.leds.set_all(False)
        if not success:
            return success

        logger.info(f"track_window: {self.track_window}")

        self.tracker.start(
            frame=image,
            track_window=self.track_window,
        )

        return True

    def update(self) -> bool:
        if not super().update():
            return False

        mode = self.config.get("mode")
        if mode == OperationMode.TRAINING:
            return self.update_training()

        if self.setpoint.speed <= 0:
            return True

        if mode == OperationMode.ACTION:
            return self.update_action()

        return True

    def update_action(self) -> bool:
        if not self.tracking_timer.tick():
            return True

        self.leds.flash("red")

        success, image = camera.capture(
            close_after=False,
            open_before=False,
            log=True,
        )
        if not success:
            return success

        debug_mode = self.config.get("debug_mode")
        _, self.track_window, output_image = self.tracker.track(
            frame=image,
            track_window=self.track_window,
            log=debug_mode,
        )
        logger.info(f"🎯 {self.track_window}")

        if debug_mode:
            if not self.send_debug_data(
                SocketMessage(
                    {
                        "image": output_image,
                    }
                )
            ):
                logger.warning("failed to send debug data.")

        if not self.tracker.tracking:
            logger.warning("🎯 lost target")
            self.leds.flash_all()
            return True

        x, _, w, _ = self.track_window
        if x + w // 2 > image.shape[1] * 2 / 3:
            self.setpoint.put(
                what="steering",
                value=-env.BLUER_UGV_SWALLOW_STEERING_SETPOINT,
                log=True,
            )
        elif x + w // 2 < image.shape[1] * 1 / 3:
            self.setpoint.put(
                what="steering",
                value=env.BLUER_UGV_SWALLOW_STEERING_SETPOINT,
                log=True,
            )

        return True

    def update_training(self) -> bool:
        self.config.set("mode", OperationMode.NONE)
        return self.select_target()



from bluer_ugv.swallow.session.classical.leds import ClassicalLeds
from bluer_ugv.swallow.session.classical.setpoint.classes import ClassicalSetPoint


class ClassicalVoidMotor:
    def __init__(
        self,
        setpoint: ClassicalSetPoint,
        leds: ClassicalLeds,
    ):
        pass

    def cleanup(self):
        pass

    def initialize(self) -> bool:
        return True

    def update(self) -> bool:
        return True



dict_of_parts = {
    "ac-volt-meter": "220 V, 5 A",
    "dc-circuit-breaker": "40 A",
    "dsn-vc288": "with 50 A shunt",
    "hdmi-cable": "",
    "mcb": "2 - 3 A @ 220 V AC",
    "micro-hdmi-adapter": "",
    "power-adapter": "12V DC, 10 A",
    "power-inverter": "continuous output ≥ 400-500 W, surge output ≥ 800 W",
    "relay": "12 V DC, 25 A",
    "safety-fuse": "30 mA trip",
    "SLA-Battery": "12 V, 42 Ah, 185 mm x 150 mm x 175 mm",
    "TV-bracket": "2 x",
    "TV": "2 x 40 inch",
}



from typing import Any

from bluer_ugv.README.ugvs.comparison.features.classes import Feature


class RangeFeature(Feature):
    nickname = "range"
    long_name = "شعاع عملکرد عملیاتی"

    @property
    def score_as_str_(self) -> str:
        return f"{self.score} کیلومتر"



from RPi import GPIO  # type: ignore
import threading

from bluer_ugv.logger import logger


class ClassicalLeds:
    def __init__(self):
        self._leds = {
            "yellow": {"pin": 17, "state": True},
            "red": {"pin": 27, "state": False},
            "green": {"pin": 22, "state": True},
        }

        self._lock = threading.Lock()

        logger.info(
            "{}: {}.".format(
                self.__class__.__name__,
                ", ".join(
                    [
                        "{}:GPIO#{}".format(
                            led_name,
                            led_info["pin"],
                        )
                        for led_name, led_info in self._leds.items()
                    ]
                ),
            )
        )

    def flash(self, color: str):
        with self._lock:
            self._leds[color]["state"] = not self._leds[color]["state"]

    def flash_all(self):
        with self._lock:
            for led in self._leds.values():
                led["state"] = not led["state"]

    def initialize(self) -> bool:
        try:
            for led in self._leds.values():
                GPIO.setup(
                    led["pin"],
                    GPIO.OUT,
                )
        except Exception as e:
            logger.error(e)
            return False

        return True

    def get(self, color: str) -> bool:
        with self._lock:
            return self._leds[color]["state"]

    def set(self, color: str, value: bool):
        with self._lock:
            self._leds[color]["state"] = value

    def set_all(
        self,
        state: bool = True,
    ) -> bool:
        with self._lock:
            for led in self._leds.values():
                led["state"] = state

        return self.update(flash_green=False)

    def update(
        self,
        flash_green: bool = True,
    ) -> bool:
        with self._lock:
            if flash_green:
                self._leds["green"]["state"] = not self._leds["green"]["state"]

            for led in self._leds.values():
                GPIO.output(
                    led["pin"],
                    GPIO.HIGH if led["state"] else GPIO.LOW,
                )

        return True



from typing import List

from bluer_options.terminal import show_usage, xtra

from bluer_ugv import env
from bluer_ugv.help.swallow.video.playlist import help_functions as help_playlist
from bluer_ugv.swallow.session.classical.screen.video.engine import VideoEngine


def help_play(
    tokens: List[str],
    mono: bool,
) -> str:
    args = [
        "[--dryrun 1]",
        "[--download 0]",
        "[--engine {}]".format(
            " | ".join(sorted([engine.name.lower() for engine in VideoEngine]))
        ),
        "[--loop 0]",
        f"[--object_name <{env.RANGIN_VIDEO_LIST_OBJECT}>]",
        "[--timeout <-1 | 10>]",
        "[--video <loading|1>]",
    ]

    return show_usage(
        [
            "@swallow",
            "video",
            "play",
        ]
        + args,
        "play <object-name>/<video>.",
        mono=mono,
    )


help_functions = {
    "play": help_play,
    "playlist": help_playlist,
}



from bluer_objects.README.items import Items

from bluer_ugv.README.rangin.consts import rangin_mechanical_design


docs = [
    {
        "path": "../docs/rangin/mechanical.md",
        "cols": 2,
        "items": Items(
            [
                {
                    "name": "arzhang",
                    "marquee": f"{rangin_mechanical_design}/robot.png?raw=true",
                    "url": f"{rangin_mechanical_design}/robot.stl",
                },
                {
                    "name": "90",
                    "marquee": f"{rangin_mechanical_design}/robot-90.png?raw=true",
                    "url": f"{rangin_mechanical_design}/robot-90.stl",
                },
                {
                    "name": "90 (without the cage)",
                    "marquee": f"{rangin_mechanical_design}/robot-90-2.png?raw=true",
                    "url": f"{rangin_mechanical_design}/robot-90.stl",
                },
                {
                    "name": "90c (curved)",
                    "marquee": f"{rangin_mechanical_design}/robot-90c.png?raw=true",
                    "url": f"{rangin_mechanical_design}/robot-90c.stl",
                },
            ]
        ),
    },
]



import argparse

from blueness import module
from blueness.argparse.generic import sys_exit

from bluer_ugv import NAME
from bluer_ugv.swallow.dataset.combination import combine
from bluer_ugv.logger import logger

NAME = module.name(__file__, NAME)

parser = argparse.ArgumentParser(NAME)
parser.add_argument(
    "task",
    type=str,
    help="combine",
)
parser.add_argument(
    "--count",
    type=int,
    default=-1,
)
parser.add_argument(
    "--download",
    type=int,
    default=1,
    help="0 | 1",
)
parser.add_argument(
    "--object_name",
    type=str,
)
parser.add_argument(
    "--recent",
    type=int,
    default=1,
    help="0 | 1",
)
parser.add_argument(
    "--split",
    type=int,
    default=1,
    help="0 | 1",
)
parser.add_argument(
    "--test_ratio",
    type=float,
    default=0.1,
)
parser.add_argument(
    "--train_ratio",
    type=float,
    default=0.8,
)
parser.add_argument(
    "--datasets",
    type=str,
    default="not-given",
    help="<object-name-1>,<object-name-2>",
)
parser.add_argument(
    "--sequence",
    type=int,
    default=-1,
)
args = parser.parse_args()

success = False
if args.task == "combine":
    success = combine(
        object_name=args.object_name,
        count=args.count,
        download=args.download == 1,
        recent=args.recent == 1,
        sequence=args.sequence,
        split=args.split == 1,
        test_ratio=args.test_ratio,
        train_ratio=args.train_ratio,
        explicit_dataset_object_names=args.datasets,
    )
else:
    success = None
sys_exit(logger, NAME, args.task, success)



import threading
from typing import Any, Dict

from bluer_ugv.swallow.session.classical.mode import OperationMode
from bluer_ugv.logger import logger


class ClassicalConfig:
    def __init__(self):
        self._lock = threading.Lock()
        self.config: Dict[str, Any] = {
            "audio_enabled": False,
            "debug_mode": False,
            "mode": OperationMode.NONE,
            "ultrasound_enabled": True,
            "ethernet_verbose": False,
        }

        logger.info(f"🎛️ created {self.as_str()}")

    def as_str(self) -> str:
        with self._lock:
            return "{}[ {} ]".format(
                self.__class__.__name__,
                ", ".join(
                    [f"{keyword}={value}" for keyword, value in self.config.items()]
                ),
            )

    def get(
        self,
        keyword: str,
        default: Any = None,
    ) -> Any:
        with self._lock:
            return self.config.get(keyword, default)

    def set(
        self,
        keyword: str,
        value: Any,
    ):
        logger.info(f"🎛️ config: {keyword} = {value}")

        with self._lock:
            self.config[keyword] = value



from bluer_objects.README.items import ImageItems

from bluer_ugv.README.arzhang.consts import arzhang_mechanical_design, arzhang_assets2

docs = [
    {
        "path": "../docs/arzhang/design",
    },
    {
        "path": "../docs/arzhang/design/specs.md",
    },
    {
        "path": "../docs/arzhang/design/power.md",
    },
    {
        "path": "../docs/arzhang/design/mechanical",
        "cols": 2,
        "items": ImageItems(
            {
                f"{arzhang_mechanical_design}/robot-with-cover-v2.png": f"{arzhang_mechanical_design}/robot.stl",
                f"{arzhang_mechanical_design}/robot.png": f"{arzhang_mechanical_design}/robot.stl",
                f"{arzhang_mechanical_design}/cage.png": f"{arzhang_mechanical_design}/cage.stl",
                f"{arzhang_mechanical_design}/measurements.png": "",
                f"{arzhang_mechanical_design}/options.png": f"{arzhang_mechanical_design}/options.svg",
                f"{arzhang_assets2}/20251128_122828-2.jpg": "",
                f"{arzhang_assets2}/20251128_155615.jpg": "",
                f"{arzhang_assets2}/20251130_134005.jpg": "",
                f"{arzhang_assets2}/20251130_140054.jpg": "",
                f"{arzhang_assets2}/20251201_132846.jpg": "",
            }
        ),
    },
    {
        "path": "../docs/arzhang/design/mechanical/v1.md",
        "items": ImageItems(
            {
                f"{arzhang_mechanical_design}/v1/robot.png": f"{arzhang_mechanical_design}/v1/robot.stl",
                f"{arzhang_mechanical_design}/v1/cage.png": f"{arzhang_mechanical_design}/v1/cage.stl",
                f"{arzhang_mechanical_design}/v1/measurements.png": "",
            }
        ),
    },
]



from bluer_ugv.README.rangin import (
    mechanical,
    operation,
    parts,
    power,
    root,
    schematics,
    specs,
    terraform,
)

docs = (
    mechanical.docs
    + operation.docs
    + parts.docs
    + power.docs
    + root.docs
    + schematics.docs
    + specs.docs
    + terraform.docs
)



from bluer_options.host.functions import is_headless

from bluer_sbc.env import BLUER_SBC_ENABLE_SCREEN

from bluer_ugv import env
from bluer_ugv.swallow.session.classical.screen.video.player import VideoPlayer
from bluer_ugv.swallow.session.classical.screen.video.playlist import PlayList
from bluer_ugv.logger import logger


class ClassicalScreen:
    def __init__(self):
        self.video_player = (
            None if (is_headless() or BLUER_SBC_ENABLE_SCREEN == 0) else VideoPlayer()
        )

        self.playlist = PlayList(env.RANGIN_VIDEO_LIST_OBJECT)

        logger.info(f"{self.__class__.__name__} created.")

    def cleanup(self):
        logger.info(f"{self.__class__.__name__}.cleanup")
        if self.video_player is not None:
            self.video_player.stop()

    def initialize(self) -> bool:
        if self.video_player is None:
            return True

        return self.video_player.play(
            self.playlist.get("loading"),
            loop=True,
        )

    def update(self) -> bool:
        if self.video_player is None:
            return True

        if self.video_player.process:
            return True

        self.playlist.next()

        return self.video_player.play(
            self.playlist.get(self.playlist.index),
            loop=False,
        )



from bluer_objects.README.items import ImageItems

from bluer_sbc.README.designs.swallow.consts import swallow_assets2

items = ImageItems(
    {
        f"{swallow_assets2}/20251113_205142.jpg": "",
        f"{swallow_assets2}/20251113_210730.jpg": "",
        f"{swallow_assets2}/20251113_210706.jpg": "",
    },
)

docs = [
    {
        "path": "../docs/swallow/digital/design/computer/shield/testing.md",
        "items": items,
    },
]



from typing import Dict, Type, List

from bluer_ugv.README.ugvs.comparison.features.classes import Feature
from bluer_ugv.README.ugvs.comparison.features.concealment import ConcealmentFeature
from bluer_ugv.README.ugvs.comparison.features.control import ControlFeature
from bluer_ugv.README.ugvs.comparison.features.cost import CostFeature
from bluer_ugv.README.ugvs.comparison.features.DYI import DYIFeature
from bluer_ugv.README.ugvs.comparison.features.payload import PayloadFeature
from bluer_ugv.README.ugvs.comparison.features.ps import PSFeature
from bluer_ugv.README.ugvs.comparison.features.range import RangeFeature
from bluer_ugv.README.ugvs.comparison.features.sanction_proof import (
    SanctionProofFeature,
)
from bluer_ugv.README.ugvs.comparison.features.size import SizeFeature
from bluer_ugv.README.ugvs.comparison.features.speed import SpeedFeature
from bluer_ugv.README.ugvs.comparison.features.swarm import SwarmFeature
from bluer_ugv.README.ugvs.comparison.features.uv_delivery import UVDeliveryFeature

list_of_feature_classes: List[Type[Feature]] = [
    PayloadFeature,
    SpeedFeature,
    SizeFeature,
    CostFeature,
    ControlFeature,
    RangeFeature,
    DYIFeature,
    SanctionProofFeature,
    UVDeliveryFeature,
    ConcealmentFeature,
    PSFeature,
    SwarmFeature,
]

dict_of_feature_classes: Dict[str, Type[Feature]] = {
    feature_class.nickname: feature_class for feature_class in list_of_feature_classes
}



from bluer_objects.README.consts import assets_url, designs_url


swallow_assets2 = assets_url(
    "swallow",
    volume=2,
)

swallow_designs = designs_url("swallow")

swallow_electrical_designs = designs_url("swallow/electrical")

swallow_mechanical_designs = designs_url("swallow/mechanical")

swallow_ultrasonic_sensor_designs = designs_url("swallow/ultrasonic-sensors")



from typing import Tuple

from bluer_ugv.README.ugvs.db import dict_of_ugvs
from bluer_ugv.logger import logger


def get_location(
    hostname: str,
    log: bool = True,
) -> Tuple[bool, str]:
    for info in dict_of_ugvs.values():
        dict_of_computers = info.get("computers", {})
        assert isinstance(dict_of_computers, dict)

        for location, hostname_ in dict_of_computers.items():
            if hostname == hostname_:

                if log:
                    logger.info(f"{hostname}.location={location}")
                return True, location

    logger.error(f"{hostname}.location not found.")
    return False, ""



from bluer_objects.README.items import ImageItems

from bluer_ugv.README.swallow.consts import swallow_assets2, swallow_designs


items = ImageItems(
    {
        f"{swallow_assets2}/20251119_193930.jpg": "",
        f"{swallow_assets2}/20251119_193954.jpg": "",
    }
)


docs = [
    {
        "path": "../docs/swallow/digital/design/computer/power.md",
        "items": items,
    },
]



C = 343.0  # speed of sound (m/s)
CYCLE_MIN_S = 0.060  # ≥60 ms per datasheet
TRIG_PULSE_S = 30e-6  # 30 µs trigger
WAIT_HIGH_TIMEOUT_S = 0.025  # 25 ms to see rising edge
WAIT_LOW_TIMEOUT_S = 0.040  # 40 ms max pulse width



from typing import List

from bluer_options.terminal import show_usage


def help_get(
    tokens: List[str],
    mono: bool,
) -> str:
    return show_usage(
        [
            "@ugv",
            "get",
            "<ugv-name>",
            "computers.back | computers.front | computers.top | <what>",
        ],
        "get ugv info.",
        mono=mono,
    )



from bluer_objects.README.items import ImageItems

from bluer_ugv.README.swallow.consts import swallow_assets2, swallow_designs
from bluer_ugv.README.computer.shield import parts, testing, v1, v2

items = ImageItems(
    {
        f"{swallow_designs}/kicad/swallow/exports/swallow.png": f"{swallow_designs}/kicad/swallow/exports/swallow.pdf",
        f"{swallow_designs}/kicad/swallow/exports/swallow-3d.png": "",
        f"{swallow_designs}/kicad/swallow/exports/swallow-3d-back.png": "",
        f"{swallow_designs}/kicad/swallow/exports/swallow-pcb.png": "",
        f"{swallow_assets2}/20251112_085331.jpg": "",
        f"{swallow_assets2}/20251112_181047.jpg": "",
        f"{swallow_assets2}/20251112_181053.jpg": "",
        f"{swallow_assets2}/20251205_171731.jpg": "",
    }
)


docs = (
    [
        {
            "path": "../docs/swallow/digital/design/computer/shield",
            "items": items,
        },
    ]
    + parts.docs
    + testing.docs
    + v1.docs
    + v2.docs
)



import argparse

from blueness import module
from blueness.argparse.generic import sys_exit

from bluer_ugv import env
from bluer_ugv import NAME
from bluer_ugv.swallow.session.classical.ethernet.testing import test
from bluer_ugv.logger import logger

NAME = module.name(__file__, NAME)

parser = argparse.ArgumentParser(NAME)
parser.add_argument(
    "task",
    type=str,
    help="test",
)
parser.add_argument(
    "--is_server",
    type=int,
    default=1,
    help="0 | 1",
)
parser.add_argument(
    "--server_name",
    type=str,
    default="0.0.0.0",
    help="0.0.0.0 | <server_name>.local",
)
args = parser.parse_args()

success = False
if args.task == "test":
    success = test(
        server_name=args.server_name,
        is_server=args.is_server == 1,
    )
else:
    success = None

sys_exit(logger, NAME, args.task, success)



from typing import Any

from bluer_algo.socket.connection import SocketConnection, DEV_HOST, DEFAULT_PORT
from bluer_sbc.imager.camera import instance as camera

from bluer_ugv.swallow.session.classical.config import ClassicalConfig
from bluer_ugv.swallow.session.classical.keyboard.classes import ClassicalKeyboard
from bluer_ugv.swallow.session.classical.leds import ClassicalLeds
from bluer_ugv.swallow.session.classical.setpoint.classes import ClassicalSetPoint
from bluer_ugv.swallow.session.classical.camera.detection import Detection


class ClassicalCamera:
    def __init__(
        self,
        config: ClassicalConfig,
        keyboard: ClassicalKeyboard,
        leds: ClassicalLeds,
        setpoint: ClassicalSetPoint,
        object_name: str,
    ):
        self.config = config
        self.keyboard = keyboard
        self.leds = leds
        self.setpoint = setpoint

        self.detection = Detection()

        self.object_name = object_name

    def initialize(self) -> bool:
        return camera.open(log=True)

    def cleanup(self):
        camera.close(log=True)

    def send_debug_data(
        self,
        data: Any,
        port: int = DEFAULT_PORT,
    ) -> bool:
        socket = SocketConnection.connect_to(
            target_host=DEV_HOST,
            port=port,
        )
        return socket.send_data(data)

    # multi-threaded support
    def stop(self):
        pass

    def update(self) -> bool:
        return True



from bluer_ugv.help.swallow.dataset import help_functions as help_dataset
from bluer_ugv.help.swallow.debug import help_debug
from bluer_ugv.help.swallow.env import help_functions as help_env
from bluer_ugv.help.swallow.ethernet import help_functions as help_ethernet
from bluer_ugv.help.swallow.git import help_functions as help_git
from bluer_ugv.help.swallow.keyboard import help_functions as help_keyboard
from bluer_ugv.help.swallow.select_target import help_select_target
from bluer_ugv.help.swallow.ultrasonic_sensor import (
    help_functions as help_ultrasonic_sensor,
)
from bluer_ugv.help.swallow.video.functions import help_functions as help_video

help_functions = {
    "dataset": help_dataset,
    "debug": help_debug,
    "env": help_env,
    "ethernet": help_ethernet,
    "git": help_git,
    "keyboard": help_keyboard,
    "select_target": help_select_target,
    "ultrasonic": help_ultrasonic_sensor,
    "video": help_video,
}



import os
from bluer_options.env import load_config, load_env, get_env

load_env(__name__)
load_config(__name__)

BLUER_UGV_MOUSEPAD_ENABLED = get_env(
    "BLUER_UGV_MOUSEPAD_ENABLED",
    0,
)

BLUER_UGV_BEAST_MODEL = get_env(
    "BLUER_UGV_BEAST_MODEL",
)

BLUER_UGV_CAMERA_ACTION_PERIOD = get_env(
    "BLUER_UGV_CAMERA_ACTION_PERIOD",
    1.5,
)
BLUER_UGV_CAMERA_TRAINING_PERIOD = get_env(
    "BLUER_UGV_CAMERA_TRAINING_PERIOD",
    3.0,
)

BLUER_UGV_SWALLOW_NAVIGATION_DATASET_LIST = get_env(
    "BLUER_UGV_SWALLOW_NAVIGATION_DATASET_LIST"
)

BLUER_UGV_SWALLOW_NAVIGATION_MODEL = get_env(
    "BLUER_UGV_SWALLOW_NAVIGATION_MODEL",
)

BLUER_UGV_SWALLOW_STEERING_SETPOINT = get_env(
    "BLUER_UGV_SWALLOW_STEERING_SETPOINT",
    50,
)

BLUER_UGV_SWALLOW_STEERING_EXPIRY = get_env(
    "BLUER_UGV_SWALLOW_STEERING_EXPIRY",
    0.05,
)

BLUER_UGV_SWALLOW_STEERING_EXPIRY_YOLO = get_env(
    "BLUER_UGV_SWALLOW_STEERING_EXPIRY_YOLO",
    0.1,
)

BLUER_UGV_SWALLOW_STEERING_EXPIRY_ETHERNET = get_env(
    "BLUER_UGV_SWALLOW_STEERING_EXPIRY_ETHERNET",
    0.3,
)

BLUER_UGV_SWALLOW_YOLO_SPEED_SETPOINT = get_env(
    "BLUER_UGV_SWALLOW_YOLO_SPEED_SETPOINT",
    20,
)

BLUER_UGV_SWALLOW_YOLO_DATASET_LIST = get_env(
    "BLUER_UGV_SWALLOW_YOLO_DATASET_LIST",
)

BLUER_UGV_SWALLOW_YOLO_MODEL = get_env(
    "BLUER_UGV_SWALLOW_YOLO_MODEL",
)

BLUER_UGV_ULTRASONIC_SENSOR_DANGER_THRESHOLD = get_env(
    "BLUER_UGV_ULTRASONIC_SENSOR_DANGER_THRESHOLD",
    500.0,
)
BLUER_UGV_ULTRASONIC_SENSOR_WARNING_THRESHOLD = get_env(
    "BLUER_UGV_ULTRASONIC_SENSOR_WARNING_THRESHOLD",
    800.0,
)

BLUER_UGV_ULTRASONIC_SENSOR_TEST_OBJECT = get_env(
    "BLUER_UGV_ULTRASONIC_SENSOR_TEST_OBJECT"
)

BLUER_UGV_ULTRASONIC_SENSOR_ENABLED = get_env(
    "BLUER_UGV_ULTRASONIC_SENSOR_ENABLED",
    1,
)

BLUER_UGV_ULTRASONIC_SENSOR_MAX_M = get_env(
    "BLUER_UGV_ULTRASONIC_SENSOR_",
    0.80,
)

BLUER_UGV_ULTRASONIC_SENSOR_KEEP_LOG = get_env(
    "BLUER_UGV_ULTRASONIC_SENSOR_KEEP_LOG",
    1,
)

BLUER_UGV_ULTRASONIC_SENSOR_LOG = get_env(
    "BLUER_UGV_ULTRASONIC_SENSOR_LOG",
    1,
)

RANGIN_VIDEO_LIST_OBJECT = get_env(
    "RANGIN_VIDEO_LIST_OBJECT",
)

BLUER_UGV_RELEASE_2 = get_env(
    "BLUER_UGV_RELEASE_2",
)

BLUER_UGV_AUDIO_LANGUAGE = get_env(
    "BLUER_UGV_AUDIO_LANGUAGE",
    "fa",
)
BLUER_UGV_AUDIO_CHANNELS = get_env(
    "BLUER_UGV_AUDIO_CHANNELS",
    1,
)
BLUER_UGV_AUDIO_RATE = get_env(
    "BLUER_UGV_AUDIO_RATE",
    48000,
)
BLUER_UGV_AUDIO_LENGTH = get_env(
    "BLUER_UGV_AUDIO_LENGTH",
    30,
)

BLUER_UGV_ETHERNET_PORT = get_env(
    "BLUER_UGV_ETHERNET_PORT",
    5050,
)

BLUER_UGV_SWALLOW_SCENARIO = get_env(
    "BLUER_UGV_SWALLOW_SCENARIO",
    "swallow",
)



from typing import List
import matplotlib

matplotlib.use("Agg")
import matplotlib.pyplot as plt
from tqdm import tqdm
import numpy as np
import cv2

from bluer_objects.graphics.signature import justify_text
from bluer_objects import objects
from bluer_objects import file, path
from bluer_objects.graphics.gif import generate_animated_gif
from bluer_objects.graphics.signature import add_signature

from bluer_ugv import env
from bluer_ugv.swallow.session.classical.ultrasonic_sensor.detection import Detection
from bluer_ugv.swallow.session.classical.ultrasonic_sensor.detection_list import (
    DetectionList,
)
from bluer_ugv.host import signature
from bluer_ugv.logger import logger


class UltrasonicSensorDetectionLog:
    def __init__(self):
        self.log: List[DetectionList] = []

    def append(self, detection_list: DetectionList):
        self.log.append(detection_list)

    def export(
        self,
        object_name: str,
        export_gif: bool = False,
        frame_count: int = -1,
        line_width: int = 80,
        log: bool = True,
        rm_blank: bool = True,
        max_m: float = env.BLUER_UGV_ULTRASONIC_SENSOR_MAX_M,
    ) -> bool:
        for func, name, filename in zip(
            [
                lambda detection: int(detection.detection),
                lambda detection: int(detection.echo_detected),
                lambda detection: detection.pulse_ms,
                lambda detection: detection.distance_mm,
            ],
            [
                "detection",
                "echo detection",
                "pulse (ms)",
                "distance(mm)",
            ],
            [
                "ultrasonic-sensor-detection",
                "ultrasonic-sensor-echo-detection",
                "ultrasonic-sensor-pulse-ms",
                "ultrasonic-sensor-distance-mm",
            ],
        ):
            plt.figure(figsize=(10, 5))

            if "distance" in name:
                plt.plot(
                    [0, len(self.log) - 1],
                    2 * [env.BLUER_UGV_ULTRASONIC_SENSOR_WARNING_THRESHOLD],
                    color="yellow",
                    linestyle=":",
                    label="warning threshold",
                )
                plt.plot(
                    [0, len(self.log) - 1],
                    2 * [env.BLUER_UGV_ULTRASONIC_SENSOR_DANGER_THRESHOLD],
                    color="red",
                    linestyle=":",
                    label="danger threshold",
                )

            plt.plot(
                [func(detection_list[0]) for detection_list in self.log],
                color="green",
                label="left sensor",
            )
            plt.plot(
                [func(detection_list[1]) for detection_list in self.log],
                color="blue",
                label="right sensor",
            )

            plt.xlim([0, len(self.log) - 1])
            if "distance" in name:
                plt.ylim([0, max_m * 1000])

            plt.title(
                justify_text(
                    " | ".join(
                        [
                            "ultrasonic-sensor",
                            name,
                        ]
                        + objects.signature(object_name=object_name)
                    ),
                    line_width=line_width,
                    return_str=True,
                )
            )
            plt.xlabel(
                justify_text(
                    " | ".join(signature()),
                    line_width=line_width,
                    return_str=True,
                )
            )
            plt.ylabel(name)
            plt.legend()
            plt.tight_layout()
            plt.grid(True)
            if not file.save_fig(
                objects.path_of(
                    object_name=object_name,
                    filename=f"{filename}.png",
                ),
                log=log,
            ):
                return False

        if not self.export_state(
            object_name=object_name,
            line_width=line_width,
            log=log,
            max_m=max_m,
        ):
            return False

        if export_gif:
            if not self.export_gif(
                object_name=object_name,
                frame_count=frame_count,
                line_width=line_width,
                log=log,
                max_m=max_m,
                rm_blank=rm_blank,
            ):
                return False

        return True

    def export_state(
        self,
        object_name: str,
        line_width: int = 80,
        height=512,
        max_m: float = env.BLUER_UGV_ULTRASONIC_SENSOR_MAX_M,
        log: bool = True,
    ) -> bool:
        image = np.zeros((len(self.log[0]), len(self.log), 3), dtype=np.uint8)

        for detection_index, detection_list in enumerate(self.log):
            for sensor_index, detection in enumerate(detection_list):
                assert isinstance(detection, Detection)

                for channel in range(3):
                    image[sensor_index, detection_index, channel] = (
                        detection.state.color_code[channel]
                    )

        image = cv2.resize(
            image,
            (
                int(image.shape[1]),
                int(image.shape[0] * (height / image.shape[0])),
            ),
            interpolation=cv2.INTER_NEAREST_EXACT,
        )

        image = add_signature(
            image,
            header=[
                " | ".join(
                    [
                        "ultrasonic sensor state",
                    ]
                    + objects.signature(object_name=object_name)
                ),
            ],
            footer=[" | ".join(signature())],
            line_width=line_width,
        )

        return file.save_image(
            objects.path_of(
                filename="ultrasonic-sensor-state.png",
                object_name=object_name,
            ),
            image,
            log=log,
        )

    def export_gif(
        self,
        object_name: str,
        line_width: int = 80,
        frame_count: int = -1,
        height: int = 512,
        width: int = 512,
        max_m: float = env.BLUER_UGV_ULTRASONIC_SENSOR_MAX_M,
        log: bool = True,
        rm_blank: bool = True,
    ) -> bool:
        image_list: List[str] = []

        rm_blank_count: int = 0

        temp_folder = objects.path_of(
            object_name=object_name,
            filename="frames",
        )
        if not path.create(temp_folder):
            return False

        for index, detection_list in tqdm(enumerate(self.log)):
            if frame_count != -1 and len(image_list) >= frame_count:
                rm_blank_count += 1
                break

            if rm_blank and all(detection.is_blank for detection in detection_list):
                continue

            filename = objects.path_of(
                object_name=object_name,
                filename=f"frames/{index:010d}.png",
            )

            image = np.concatenate(
                [
                    detection.as_image(
                        height=height,
                        width=int(width / len(detection_list)),
                        max_m=max_m,
                        line_width=int(line_width / len(detection_list)),
                        sign=False,
                    )
                    for detection in detection_list
                ],
                axis=1,
            )

            image = add_signature(
                image,
                header=[
                    " | ".join(
                        [
                            "ultrasonic-sensor",
                            "max distance: {:.2f} mm".format(max_m * 1000),
                            "warning: {:.2f} mm".format(
                                env.BLUER_UGV_ULTRASONIC_SENSOR_WARNING_THRESHOLD
                            ),
                            "danger: {:.2f} mm".format(
                                env.BLUER_UGV_ULTRASONIC_SENSOR_DANGER_THRESHOLD
                            ),
                        ]
                        + detection_list.as_str(short=True)
                        + objects.signature(
                            "frame #{:04d}/{}".format(
                                index,
                                len(self.log),
                            ),
                            object_name,
                        ),
                    )
                ],
                footer=[" | ".join(signature())],
                line_width=line_width,
            )

            if not file.save_image(filename, image):
                return False

            image_list.append(filename)

        if not generate_animated_gif(
            image_list,
            objects.path_of(
                filename="ultrasonic-sensor-detections.gif",
                object_name=object_name,
            ),
            log=log,
        ):
            return False

        if rm_blank:
            logger.info("removed {} blank frame(s).".format(rm_blank_count))

        return path.delete(temp_folder)

    def load(
        self,
        object_name: str,
    ) -> bool:
        success, self.log = file.load(
            objects.path_of(
                object_name=object_name,
                filename="detections.dill",
            ),
        )

        if success:
            logger.info("loaded {} detection(s).".format(len(self.log)))

        return success

    def save(
        self,
        object_name: str,
        log: bool = True,
    ) -> bool:
        if not file.save_yaml(
            objects.path_of(
                object_name=object_name,
                filename="detections.yaml",
            ),
            {
                "detections": [
                    [detection.as_dict() for detection in detection_list]
                    for detection_list in self.log
                ]
            },
            log=log,
        ):
            return False

        if not file.save(
            objects.path_of(
                object_name=object_name,
                filename="detections.dill",
            ),
            self.log,
            log=log,
        ):
            return False

        return True



from bluer_objects.README.alias import list_of_aliases

from bluer_ugv import NAME

docs = [
    {
        "path": "../docs/aliases",
        "macros": {
            "aliases:::": list_of_aliases(
                NAME,
                itemized=True,
            ),
        },
    }
] + [
    {
        "path": f"../docs/aliases/{alias_name}.md",
    }
    for alias_name in list_of_aliases(
        NAME,
        as_markdown=False,
    )
]



from bluer_objects.README.items import ImageItems

from bluer_ugv.README.rangin.consts import rangin_electrical_design

docs = [
    {
        "path": "../docs/rangin/schematics.md",
        "items": ImageItems(
            {
                f"{rangin_electrical_design}/electrical.png?raw=true": f"{rangin_electrical_design}/electrical.svg",
            },
        ),
    },
]



from bluer_objects.README.items import ImageItems

from bluer_ugv.README.rangin.consts import rangin_assets2
from bluer_ugv.README.ugvs.db import dict_of_ugvs

items = ImageItems(
    {
        **{
            f"{rangin_assets2}/concepts/11.png": "",
            f"{rangin_assets2}/concepts/12.png": "",
        },
        **{url: "" for url in dict_of_ugvs["rangin"]["items"]},
    }
)



import argparse
import time

from blueness import module
from blueness.argparse.generic import sys_exit
from bluer_options.host import is_rpi, is_headless
from bluer_options import string

from bluer_ugv import NAME, env
from bluer_ugv.swallow.session.classical.screen.video.playlist import PlayList
from bluer_ugv.swallow.session.classical.screen.video.player import VideoPlayer
from bluer_ugv.swallow.session.classical.screen.video.engine import VideoEngine
from bluer_ugv.logger import logger

NAME = module.name(__file__, NAME)

parser = argparse.ArgumentParser(NAME)
parser.add_argument(
    "task",
    type=str,
    help="play",
)
parser.add_argument(
    "--download",
    type=int,
    default=1,
    help="0|1",
)
parser.add_argument(
    "--dryrun",
    type=int,
    default=int(not is_rpi() or is_headless()),
    help="0|1",
)
parser.add_argument(
    "--engine",
    type=str,
    default=VideoEngine.VLC.name.lower(),
    help=" | ".join(sorted([engine.name.lower() for engine in VideoEngine])),
)
parser.add_argument(
    "--loop",
    type=int,
    default=1,
    help="0|1",
)
parser.add_argument(
    "--object_name",
    type=str,
    default=env.RANGIN_VIDEO_LIST_OBJECT,
)
parser.add_argument(
    "--timeout",
    type=int,
    default=-1,
    help="in seconds, -1: never",
)
parser.add_argument(
    "--video",
    type=str,
    default="loading",
)
args = parser.parse_args()


success = False
if args.task == "play":
    playlist = PlayList(
        args.object_name,
        download=args.download == 1,
    )

    video_player = VideoPlayer(
        args.dryrun == 1,
        engine=VideoEngine[args.engine.upper()],
    )

    success = video_player.play(
        filename=playlist.get(args.video),
        loop=args.loop == 1,
    )

    if success and args.timeout > 0:
        logger.info(
            "waiting for {}".format(
                string.pretty_duration(
                    args.timeout,
                )
            )
        )
        time.sleep(args.timeout)
        success = video_player.stop()

        logger.info('💡 type in "reset" if the prompt is invisible.')

else:
    success = None

sys_exit(logger, NAME, args.task, success)



import keyboard

from blueness import module

from bluer_ugv import NAME
from bluer_ugv import env
from bluer_ugv.swallow.session.classical.ethernet.client import EthernetClient
from bluer_ugv.swallow.session.classical.ethernet.command import EthernetCommand
from bluer_ugv.logger import logger


NAME = module.name(__file__, NAME)


def test(
    server_name: str,
    is_server: bool,
    port: int = env.BLUER_UGV_ETHERNET_PORT,
) -> bool:
    success = True

    logger.info(
        "{}.test: server_name={}, is_server={}, port={}".format(
            NAME,
            server_name,
            is_server,
            port,
        )
    )

    client = EthernetClient(
        host=server_name,
        port=port,
        is_server=is_server,
    )

    logger.info("press 5 to send a message.")
    try:
        while True:
            client.process()

            if keyboard.is_pressed("5"):
                client.send(EthernetCommand(action="hello"))
    except KeyboardInterrupt:
        logger.info("Ctrl+C, stopping.")
    except Exception as e:
        logger.error(e)
        success = False

    client.close()

    return success



import threading

from bluer_options.env import abcli_hostname

from bluer_ugv import env
from bluer_ugv.swallow.session.classical.config import ClassicalConfig
from bluer_ugv.swallow.session.classical.ethernet.client import EthernetClient
from bluer_ugv.README.ugvs.ethernet import find_server
from bluer_ugv.logger import logger


class ClassicalEthernet:
    def __init__(
        self,
        config: ClassicalConfig,
    ):
        self.config = config

        self.enabled: bool = True

        logger.info(f"creating {self.__class__.__name__}...")

        self.running = False

        self.enabled, is_server, server_name = find_server(hostname=abcli_hostname)
        if not self.enabled:
            return

        self.client = EthernetClient(
            host=server_name,
            port=env.BLUER_UGV_ETHERNET_PORT,
            is_server=is_server,
            verbose=self.config.get("ethernet_verbose"),
        )

        self.running = True
        self.thread = threading.Thread(target=self.loop, daemon=True)
        self.thread.start()

    def stop(self):
        if not self.enabled:
            return

        self.running = False
        self.thread.join()

        self.client.close()
        logger.info(f"{self.__class__.__name__}.stopped.")

    def loop(self):
        logger.info(f"{self.__class__.__name__}.loop started.")

        while self.running:
            self.client.process()



from filelock import FileLock

from bluer_objects import file, objects

from bluer_algo.bps.position import Position
from bluer_algo.logger import logger


class ClassicalPosition:
    def __init__(self, object_name):
        self.object_name: str = object_name
        self.updated: bool = False
        self.position: Position = Position()

        logger.info(f"{self.__class__.__name__}: {object_name}")

    def initialize(self) -> bool:
        return self.update()

    def update(self) -> bool:
        self.updated = False

        filename = objects.path_of(
            object_name=self.object_name,
            filename="position.yaml",
        )

        if not file.exists(filename):
            return True

        self.updated, self.position = Position.load(self.object_name)

        lock = FileLock(filename + ".lock")

        with lock:
            file.delete(filename)

        return True



from bluer_objects.README.consts import assets_url, designs_url

arzhang_assets = assets_url(
    suffix="arzhang",
)

arzhang_assets2 = assets_url(
    suffix="arzhang",
    volume=2,
)

arzhang2_assets2 = assets_url(
    suffix="arzhang2",
    volume=2,
)

arzhang_mechanical_design = designs_url("arzhang")



from typing import List

from bluer_options.terminal import show_usage, xtra

from bluer_ugv import env


def help_test(
    tokens: List[str],
    mono: bool,
) -> str:
    options = xtra("dryrun", mono=mono)

    args = [
        "[--keys 1234567890-+/.]",
    ]

    return show_usage(
        [
            "@swallow",
            "keyboard",
            "test",
            f"[{options}]",
        ]
        + args,
        "test keyboard.",
        mono=mono,
    )


help_functions = {
    "test": help_test,
}



from __future__ import annotations
from dataclasses import dataclass, field
from typing import Dict


@dataclass
class EthernetCommand:
    action: str = ""
    data: Dict = field(default_factory=dict)

    def as_str(self) -> str:
        return f"{self.__class__.__name__}({self.action})[{self.data}]"

    def to_dict(self) -> Dict:
        return {"action": self.action, "data": self.data}

    @staticmethod
    def from_dict(d: Dict) -> "EthernetCommand":
        return EthernetCommand(
            action=str(d.get("action", "")),
            data=dict(d.get("data", {}) or {}),
        )



from typing import List, Any, Union, Tuple, Dict
from enum import Enum, auto

from bluer_ugv.logger import logger


class Feature_Enum(Enum):
    @property
    def as_str(self):
        return self.name


class Feature_Comparison(Feature_Enum):
    SIMILAR = auto()
    HIGHER = auto()
    LOWER = auto()

    UNKNOWN = auto()


class Feature:
    nickname: str
    long_name: str

    comparison_as_str: Dict[Feature_Comparison, str] = {
        Feature_Comparison.HIGHER: "بیشتر",
        Feature_Comparison.LOWER: "کمتر",
        Feature_Comparison.SIMILAR: "مشابه",
    }

    def __init__(
        self,
        score: Any,
    ):
        self.score = score

    def compare(
        self,
        feature: Union["Feature", None],
        ugv_name: str,
        log: bool = False,
    ) -> Tuple[Feature_Comparison, str]:
        comparison: Feature_Comparison = Feature_Comparison.HIGHER
        if feature is None:
            if isinstance(self.score, bool):
                comparison = (
                    Feature_Comparison.HIGHER
                    if self.score
                    else Feature_Comparison.LOWER
                )
            else:
                comparison = Feature_Comparison.UNKNOWN
        elif feature.score is Ellipsis:
            comparison = Feature_Comparison.UNKNOWN
        elif self.score_index == feature.score_index:
            comparison = Feature_Comparison.SIMILAR
        elif self.score_index < feature.score_index:
            comparison = Feature_Comparison.LOWER

        if log:
            logger.info(
                "{}: {} ({}) vs. {} ({}): {}".format(
                    self.__class__.__name__,
                    self.score,
                    self.score_index,
                    "-" if feature is None else feature.score,
                    "-" if feature is None else feature.score_index,
                    comparison.name,
                )
            )

        return comparison, self.describe_status(
            comparison,
            ugv_name,
        )

    def describe_status(
        self,
        comparison: Feature_Comparison,
        ugv_name: str,
    ) -> str:
        return "{} {}{}{}".format(
            self.long_name,
            self.__class__.comparison_as_str.get(comparison, ""),
            (
                " در {} ".format(ugv_name)
                if comparison != Feature_Comparison.SIMILAR
                else ""
            ),
            (lambda info: f" ({info})" if info else "")(self.score_as_str()),
        )

    @property
    def description(self) -> str:
        return "{} {}".format(
            self.long_name,
            self.score_as_str(force=True),
        )

    def score_as_str(
        self,
        force: bool = False,
    ) -> str:
        return (
            self.score.as_str
            if isinstance(self.score, Feature_Enum) and force
            else self.score_as_str_
        )

    @property
    def score_as_str_(self) -> str:
        return ""

    @property
    def score_index(self):
        return (
            int(self.score)
            if isinstance(self.score, bool)
            else self.score.value if isinstance(self.score, Enum) else self.score
        )


class FeatureList:
    def __init__(self):
        self.db: List[Feature] = []

    def add(self, feature: Feature):
        self.db.append(feature)

    def get(
        self,
        feature_name: str,
    ) -> Union[Feature, None]:
        for feature in self.db:
            if feature_name == feature.nickname:
                return feature

        return None



from bluer_ugv.swallow.session.classical.camera.generic import ClassicalCamera


class ClassicalVoidCamera(ClassicalCamera):
    def initialize(self) -> bool:
        return True

    def cleanup(self):
        pass



from bluer_objects.README.items import ImageItems

from bluer_ugv.README.swallow.consts import swallow_electrical_designs


items = ImageItems(
    {
        f"{swallow_electrical_designs}/digital.png": f"{swallow_electrical_designs}/digital.svg",
    }
)

docs = [
    {
        "path": "../docs/swallow/digital/design/computer/schematics.md",
        "items": items,
    },
]



from bluer_objects.README.items import ImageItems

from bluer_ugv.README.swallow.consts import swallow_assets2

items = ImageItems(
    {
        f"{swallow_assets2}/20251116_145939.jpg": "",
        f"{swallow_assets2}/20251116_150940.jpg": "",
        f"{swallow_assets2}/20251116_151611.jpg": "",
        f"{swallow_assets2}/20251116_152801.jpg": "",
        f"{swallow_assets2}/20251116_152832_1.gif": "",
        f"{swallow_assets2}/2026-01-25-12-21.jpg": "",
        f"{swallow_assets2}/2026-01-25-12-16.jpg": "",
    }
)

docs = [
    {
        "path": "../docs/swallow/digital/design/computer/testing.md",
        "items": items,
    },
]



from typing import List

from bluer_options.terminal import show_usage, xtra
from bluer_objects.help.download import options as download_options
from bluer_objects.help.upload import options as upload_options


def help_cat(
    tokens: List[str],
    mono: bool,
) -> str:
    options = xtra("download", mono=mono)

    return show_usage(
        [
            "@swallow",
            "video",
            "playlist",
            "cat",
            f"[{options}]",
        ],
        "cat swallow playlist.",
        mono=mono,
    )


def help_download(
    tokens: List[str],
    mono: bool,
) -> str:
    options = download_options(mono=mono)

    return show_usage(
        [
            "@swallow",
            "video",
            "playlist",
            "download",
            f"[{options}]",
        ],
        "download swallow playlist.",
        mono=mono,
    )


def help_edit(
    tokens: List[str],
    mono: bool,
) -> str:
    options = xtra("download", mono=mono)

    return show_usage(
        [
            "@swallow",
            "video",
            "playlist",
            "edit",
            f"[{options}]",
        ],
        "edit swallow playlist.",
        mono=mono,
    )


def help_upload(
    tokens: List[str],
    mono: bool,
) -> str:
    options = upload_options(mono=mono)

    return show_usage(
        [
            "@swallow",
            "video",
            "playlist",
            "upload",
            f"[{options}]",
        ],
        "upload swallow playlist.",
        mono=mono,
    )


help_functions = {
    "cat": help_cat,
    "download": help_download,
    "edit": help_edit,
    "upload": help_upload,
}



from bluer_ugv.README.swallow.items import items
from bluer_ugv.README.swallow import analog, digital

docs = (
    [
        {
            "items": items,
            "path": "../docs/swallow",
        },
    ]
    + analog.docs
    + digital.docs
)



class Detection:
    def __init__(
        self,
        x: float = 0.0,
        y: float = 0.0,
        confidence: float = 0.0,
        area: float = 0.0,
    ):
        self.x: float = x
        self.y: float = y
        self.confidence: float = confidence
        self.area: float = area

    def as_str(self) -> str:
        return "{} @[{:.2f},{:.2f}](C:{:.2f}%,A:{:.2f}%)".format(
            self.__class__.__name__,
            self.x,
            self.y,
            self.confidence * 100,
            self.area * 100,
        )



from typing import List

from bluer_options.terminal import show_usage


def help_watch(
    tokens: List[str],
    mono: bool,
) -> str:
    return show_usage(
        [
            "@ugv",
            "watch",
            "<ugv-name>",
            "[back | front | top | <node>]",
        ],
        "watch <ugv-name>.<node>.",
        mono=mono,
    )



from bluer_objects.README.consts import assets_url, designs_repo


rangin_assets2 = assets_url(
    suffix="rangin",
    volume=2,
)

rangin_electrical_design = f"{designs_repo}/blob/main/rangin/electrical"

rangin_mechanical_design = f"{designs_repo}/blob/main/rangin"



from bluer_ugv.swallow.session.classical.camera.generic import (
    ClassicalCamera,
)
from bluer_ugv.swallow.session.classical.camera.navigation import (
    ClassicalNavigationCamera,
)
from bluer_ugv.swallow.session.classical.camera.tracking import (
    ClassicalTrackingCamera,
)
from bluer_ugv.swallow.session.classical.camera.void import (
    ClassicalVoidCamera,
)
from bluer_ugv.swallow.session.classical.camera.yolo import (
    ClassicalYoloCamera,
)



from typing import List


class Reference:
    def __init__(
        self,
        title: str,
        url: str,
        list_of_ugvs: List[str] = [],
        is_in_farsi: bool = False,
    ):
        self.title = title
        self.url = url
        self.list_of_ugvs = list_of_ugvs
        self.is_in_farsi = is_in_farsi


class List_of_References:
    def __init__(self):
        self.db: List[Reference] = []

    def add(
        self,
        **kw_args,
    ):
        ugv = Reference(**kw_args)
        self.db.append(ugv)



from typing import Union, Dict

from bluer_ugv import env
from bluer_ugv.swallow.session.classical.ethernet.command import EthernetCommand
from bluer_ugv.swallow.session.classical.setpoint.classes import ClassicalSetPoint


class ClassicalEthernetSetPoint(ClassicalSetPoint):
    def put(
        self,
        value: Union[int, bool, Dict[str, Union[int, bool]]],
        what: str = "all",
        log: bool = True,
        steering_expires_in: float = env.BLUER_UGV_SWALLOW_STEERING_EXPIRY_ETHERNET,
    ):
        super().put(
            value=value,
            what=what,
            log=log,
            steering_expires_in=steering_expires_in,
        )

        self.ethernet.client.send(
            EthernetCommand(
                action="setpoint.put",
                data={
                    "value": value,
                    "what": what,
                    "steering_expires_in": steering_expires_in,
                },
            )
        )



from bluer_ai.help.generic import help_functions as generic_help_functions

from bluer_ugv import ALIAS
from bluer_ugv.help.get import help_get
from bluer_ugv.help.git import help_git
from bluer_ugv.help.ssh import help_ssh
from bluer_ugv.help.swallow import help_functions as help_swallow
from bluer_ugv.help.watch import help_watch


help_functions = generic_help_functions(plugin_name=ALIAS)

help_functions.update(
    {
        "get": help_get,
        "git": help_git,
        "ssh": help_ssh,
        "swallow": help_swallow,
        "watch": help_watch,
    }
)



from bluer_objects.README.items import ImageItems

from bluer_ugv.README.swallow.consts import (
    swallow_assets2,
    swallow_electrical_designs,
)

items = ImageItems(
    {
        f"{swallow_electrical_designs}/nuts-bolts-spacers.png": f"{swallow_electrical_designs}/nuts-bolts-spacers.svg",
        f"{swallow_assets2}/20251018_133202.jpg": "",
        f"{swallow_assets2}/20251018_133349.jpg": "",
        f"{swallow_assets2}/20251008_114557.jpg": "",
        f"{swallow_assets2}/20251008_133418.jpg": "",
        f"{swallow_assets2}/20251008_124129.jpg": "",
        f"{swallow_assets2}/20251008_124932.jpg": "",
        f"{swallow_assets2}/20260113_172208.jpg": "",
    }
)

docs = [
    {
        "path": "../docs/swallow/digital/design/computer/box.md",
        "items": items,
    }
]



from typing import Dict, List

from bluer_options.logger.config import log_dict, log_list
from bluer_objects.metadata import get_from_object
from bluer_objects import storage, objects
from bluer_objects.storage.policies import DownloadPolicy

from bluer_ugv.logger import logger


class PlayList:
    def __init__(
        self,
        object_name: str,
        download: bool = True,
    ):
        self.index: int = -1

        self.object_name = object_name
        self.download = download

        if self.download:
            storage.download(
                self.object_name,
                filename="metadata.yaml",
            )

        self.messages: Dict[str, str] = get_from_object(
            self.object_name,
            "messages",
            default={},
        )
        log_dict(
            logger,
            "loaded",
            self.messages,
            "message(s)",
            max_count=-1,
            max_length=-1,
        )

        self.playlist: List[Dict[str, str]] = get_from_object(
            self.object_name,
            "playlist",
            default=[],
        )
        log_list(
            logger,
            "loaded",
            self.playlist,
            "playlist item(s)",
            max_count=-1,
            max_length=-1,
        )

        logger.info(
            "{} created from {}.".format(
                self.__class__.__name__,
                self.object_name,
            )
        )

    def get(
        self,
        keyword: int | str = "loading",
        what: str = "filename",
    ) -> str:
        filename = f"{keyword.__class__.__name__}-not-supported"

        if isinstance(keyword, int):
            filename = "bad-index-{}-from-{}".format(
                keyword,
                len(self.playlist),
            )

            if 0 <= keyword < len(self.playlist):
                filename = self.playlist[keyword].get(
                    what,
                    f"{what}-not-found",
                )

        if isinstance(keyword, str):
            filename = f"{keyword}-not-found"

            if keyword.isnumeric():
                return self.get(int(keyword), what=what)

            if keyword in self.messages:
                filename = self.messages[keyword].get(
                    what,
                    f"{what}-not-found",
                )

        if self.download:
            storage.download(
                self.object_name,
                filename=filename,
                policy=DownloadPolicy.DOESNT_EXIST,
            )

        return objects.path_of(
            filename=filename,
            object_name=self.object_name,
        )

    def next(self):
        self.index += 1
        if self.index >= len(self.playlist):
            self.index = 0

        logger.info(
            "{}: video #{}".format(
                self.__class__.__name__,
                self.index,
            )
        )



import json
from queue import Empty, SimpleQueue
import socket
from typing import Tuple, Dict, Optional
import struct
import threading
import time

from bluer_ugv.logger import logger
from bluer_ugv.swallow.session.classical.ethernet.command import EthernetCommand
from bluer_ugv.logger import logger


class EthernetClient:
    def __init__(
        self,
        host: str,
        port: int,
        is_server: bool = False,
        reconnect_sec: float = 1.0,
        verbose: bool = True,
    ):
        self.host = host
        self.port = port
        self.is_server = is_server
        self.reconnect_sec = reconnect_sec
        self.verbose = verbose

        self._send_queue: SimpleQueue[EthernetCommand] = SimpleQueue()
        self._receive_queue: SimpleQueue[EthernetCommand] = SimpleQueue()

        self._lock = threading.Lock()
        self._sock: Optional[socket.socket] = None
        self._listener: Optional[socket.socket] = None

        logger.info(
            "{} created: host={}, port={}{}{}".format(
                self.__class__.__name__,
                self.host,
                self.port,
                ", server" if is_server else "",
                ", verbose" if verbose else "",
            )
        )

    def _close_sockets(self):
        with self._lock:
            if self._sock:
                try:
                    self._sock.close()
                    logger.info(f"{self.__class__.__name__}._sock closed.")
                except Exception as e:
                    logger.warning(e)
                self._sock = None

            if self._listener:
                try:
                    self._listener.close()
                    logger.info(f"{self.__class__.__name__}._listener closed.")
                except Exception as e:
                    logger.warning(e)
                self._listener = None

    def _drain_send_queue(self) -> bool:
        with self._lock:
            sock = self._sock
        if sock is None:
            return False

        while True:
            try:
                cmd = self._send_queue.get_nowait()
            except Empty:
                return True

            try:
                payload = cmd.to_dict()
                # Sending can block; keep it short and safe
                sock.setblocking(True)
                try:
                    self._send_dict(sock, payload)
                finally:
                    sock.setblocking(False)

                if self.verbose:
                    logger.info(f"{self.__class__.__name__}: sent {cmd.as_str()}")

            except (ConnectionError, OSError) as e:
                logger.warning(f"{self.__class__.__name__}: send error: {e}")
                self._close_sockets()
                # put back? your choice; here we drop to avoid infinite resend
                return False
            except Exception as e:
                logger.error(e)
                return False

            return True

    def _ensure_connection(self) -> bool:
        """
        Ensure self._sock is a connected TCP socket.
        Returns True if connected.
        """
        with self._lock:
            if self._sock is not None:
                return True

        try:
            if self.is_server:
                # Lazy-create listener
                with self._lock:
                    if self._listener is None:
                        lst = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
                        lst.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
                        lst.bind((self.host, self.port))
                        lst.listen(1)
                        lst.settimeout(1.0)  # so thread can exit cleanly
                        self._listener = lst
                        logger.info(
                            f"{self.__class__.__name__}: listening on {self.host}:{self.port}"
                        )

                # Accept
                try:
                    conn, addr = self._listener.accept()
                except socket.timeout:
                    logger.warning("socket timeout.")
                    return False
                except Exception as e:
                    logger.error(e)
                    return False

                conn.setblocking(False)
                with self._lock:
                    self._sock = conn
                logger.info(f"{self.__class__.__name__}: accepted {addr}")

            else:
                # Client connect
                sock = socket.create_connection((self.host, self.port), timeout=2.0)
                sock.setblocking(False)
                with self._lock:
                    self._sock = sock
                logger.info(
                    f"{self.__class__.__name__}: connected to {self.host}:{self.port}"
                )

            return True

        except Exception as e:
            logger.warning(f"{self.__class__.__name__}: connection failed: {e}")
            self._close_sockets()
            return False

    def _recv_dict_blocking(
        self,
        sock: socket.socket,
    ) -> Dict:
        header = self._recv_exact(sock, 4)
        msg_len = struct.unpack("!I", header)[0]
        raw = self._recv_exact(sock, msg_len)
        return json.loads(raw.decode("utf-8"))

    @staticmethod
    def _recv_exact(
        sock: socket.socket,
        n: int,
    ) -> bytes:
        buf = b""
        while len(buf) < n:
            chunk = sock.recv(n - len(buf))
            if not chunk:
                raise ConnectionError("socket closed")
            buf += chunk
        return buf

    @staticmethod
    def _send_dict(
        sock: socket.socket,
        payload: Dict,
    ) -> None:
        raw = json.dumps(payload).encode("utf-8")
        header = struct.pack("!I", len(raw))
        sock.sendall(header + raw)

    def _try_recv_one(self) -> Tuple[bool, EthernetCommand]:
        """
        Non-blocking receive of exactly one command.
        Uses a small state machine via MSG_PEEK not needed: instead we temporarily
        switch to blocking only when we know bytes are ready.
        """
        with self._lock:
            sock = self._sock
        if sock is None:
            return False, EthernetCommand()

        try:
            # Peek header availability (4 bytes)
            try:
                hdr = sock.recv(4, socket.MSG_PEEK)
            except BlockingIOError:
                return False, EthernetCommand()

            if len(hdr) < 4:
                return False, EthernetCommand()

            msg_len = struct.unpack("!I", hdr)[0]
            total = 4 + msg_len

            try:
                blob = sock.recv(total, socket.MSG_PEEK)
            except BlockingIOError:
                return False, EthernetCommand()

            if len(blob) < total:
                return False, EthernetCommand()

            # Now actually read for real (blocking reads are safe because we already peeked)
            sock.setblocking(True)
            try:
                d = self._recv_dict_blocking(sock)
            finally:
                sock.setblocking(False)

            return True, EthernetCommand.from_dict(d)

        except (ConnectionError, OSError) as e:
            logger.warning(f"{self.__class__.__name__}: recv error: {e}")
            self._close_sockets()
            return False, EthernetCommand()
        except Exception as e:
            logger.warning(f"{self.__class__.__name__}: recv parse error: {e}")
            return False, EthernetCommand()

    def close(self):
        self._close_sockets()

    def process(self):
        connected = self._ensure_connection()
        if not connected:
            time.sleep(self.reconnect_sec)
            return

        # 1) receive at most one per tick (cheap + predictable)
        received, command = self._try_recv_one()
        if received:
            if self.verbose:
                logger.info(
                    "{} received {}".format(
                        self.__class__.__name__,
                        command.as_str(),
                    )
                )

            self._receive_queue.put(command)

        # 2) drain outbound queue
        self._drain_send_queue()

        time.sleep(0.1)

    def send(
        self,
        command: EthernetCommand,
        drain: bool = False,
    ):
        self._send_queue.put(command)

        if self.verbose:
            logger.info(
                "{}.send: queue += {}".format(
                    self.__class__.__name__,
                    command.as_str(),
                )
            )

        if drain:
            self._drain_send_queue()



from typing import List, Dict

from bluer_options.terminal import show_usage
from bluer_sbc import env

dict_of_variables: Dict[str, Dict] = {
    "bps": {
        "name": "BLUER_SBC_SWALLOW_HAS_BPS",
        "value": env.BLUER_SBC_SWALLOW_HAS_BPS,
    },
    "camera": {
        "name": "BLUER_SBC_SWALLOW_HAS_CAMERA",
        "value": env.BLUER_SBC_SWALLOW_HAS_CAMERA,
    },
    "dev_mode": {
        "name": "BLUER_SBC_SWALLOW_DEV_MODE",
        "value": env.BLUER_SBC_SWALLOW_DEV_MODE,
    },
    "full_keyboard": {
        "name": "BLUER_SBC_SWALLOW_HAS_FULL_KEYBOARD",
        "value": env.BLUER_SBC_SWALLOW_HAS_FULL_KEYBOARD,
    },
    "screen": {
        "name": "BLUER_SBC_ENABLE_SCREEN",
        "value": env.BLUER_SBC_ENABLE_SCREEN,
    },
    "steering": {
        "name": "BLUER_SBC_SWALLOW_HAS_STEERING",
        "value": env.BLUER_SBC_SWALLOW_HAS_STEERING,
    },
}


def help_cat(
    tokens: List[str],
    mono: bool,
) -> str:
    return show_usage(
        [
            "@swallow",
            "env",
            "cat",
            "[<env-name>]",
        ],
        "cat swallow-raspbian-<env-name>.env.",
        mono=mono,
    )


def help_cd(
    tokens: List[str],
    mono: bool,
) -> str:
    return show_usage(
        [
            "@swallow",
            "env",
            "cd",
        ],
        "cd env folder.",
        mono=mono,
    )


def help_cp(
    tokens: List[str],
    mono: bool,
) -> str:
    return show_usage(
        [
            "@swallow",
            "env",
            "cp",
            "[<env-name>]",
        ],
        "cp swallow-raspbian-<env-name>.env.",
        mono=mono,
    )


def help_list(
    tokens: List[str],
    mono: bool,
) -> str:
    return show_usage(
        [
            "@swallow",
            "env",
            "list",
        ],
        "list swallow envs.",
        mono=mono,
    )


def help_set(
    tokens: List[str],
    mono: bool,
) -> str:
    return show_usage(
        [
            "@swallow",
            "env",
            "set",
            " | ".join(sorted(dict_of_variables.keys())),
            "0 | 1",
        ],
        "set env.",
        {
            "{}: {} (currently: {})".format(
                keyword,
                info["name"],
                info["value"],
            ): ""
            for keyword, info in dict_of_variables.items()
        },
        mono=mono,
    )


help_functions = {
    "cat": help_cat,
    "cd": help_cd,
    "cp": help_cp,
    "list": help_list,
    "set": help_set,
}



from typing import List

from bluer_options.terminal import show_usage, xtra


def help_rm_keys(
    tokens: List[str],
    mono: bool,
) -> str:
    options = "~dryrun,undo"

    return show_usage(
        [
            "@swallow",
            "git",
            "rm_keys",
            f"[{options}]",
        ],
        "(undo) rm github keys.",
        mono=mono,
    )


help_functions = {
    "rm_keys": help_rm_keys,
}



import threading
import time

from bluer_objects.env import abcli_object_name
from bluer_sbc.env import BLUER_SBC_SWALLOW_DEV_MODE

from bluer_ugv import env
from bluer_ugv.swallow.session.classical.ultrasonic_sensor.pack import (
    UltrasonicSensorPack,
)
from bluer_ugv.swallow.session.classical.ultrasonic_sensor.detection import (
    DetectionState,
)
from bluer_ugv.swallow.session.classical.ultrasonic_sensor.detection_list import (
    DetectionList,
)
from bluer_ugv.swallow.session.classical.ultrasonic_sensor.log import (
    UltrasonicSensorDetectionLog,
)
from bluer_ugv.swallow.session.classical.config import ClassicalConfig
from bluer_ugv.swallow.session.classical.setpoint.classes import ClassicalSetPoint
from bluer_ugv.logger import logger


class ClassicalUltrasonicSensor:
    def __init__(
        self,
        config: ClassicalConfig,
        setpoint: ClassicalSetPoint,
    ):
        self.enabled = env.BLUER_UGV_ULTRASONIC_SENSOR_ENABLED == 1
        logger.info(
            "{}: {}".format(
                self.__class__.__name__,
                (
                    "enabled: warning<{:.2f} mm, danger<{:.2f} mm".format(
                        env.BLUER_UGV_ULTRASONIC_SENSOR_WARNING_THRESHOLD,
                        env.BLUER_UGV_ULTRASONIC_SENSOR_DANGER_THRESHOLD,
                    )
                    if self.enabled
                    else "disabled"
                ),
            )
        )

        self.config = config
        self.setpoint = setpoint

        self.detection_list = DetectionList()

        self.pack = None
        self.log = None
        self.running = False

        if not self.enabled:
            return

        self.pack = UltrasonicSensorPack(
            setmode=False,
            max_m=env.BLUER_UGV_ULTRASONIC_SENSOR_MAX_M,
        )

        if not self.pack.valid:
            raise NameError("valid ultrasonic sensor pack.")

        if env.BLUER_UGV_ULTRASONIC_SENSOR_KEEP_LOG == 1:
            self.log = UltrasonicSensorDetectionLog()

        self.running = True
        self.thread = threading.Thread(target=self.loop, daemon=True)
        self.thread.start()

    def stop(self):
        if not self.enabled:
            return

        self.running = False
        self.thread.join()

        logger.info(f"{self.__class__.__name__}.stopped.")

        if self.log is not None and BLUER_SBC_SWALLOW_DEV_MODE == 0:
            self.log.save(object_name=abcli_object_name)
            self.log.export(object_name=abcli_object_name)

    def loop(self):
        logger.info(f"{self.__class__.__name__}.loop started.")

        while self.running:
            if not self.config.get("ultrasound_enabled"):
                time.sleep(0.01)
                continue

            success, self.detection_list = self.pack.detect(
                log=env.BLUER_UGV_ULTRASONIC_SENSOR_LOG == 1,
            )
            if not success:
                raise NameError("failed to detect ultrasonic sensor.")

            if self.log is not None:
                self.log.append(self.detection_list)

            log_detections: bool = False
            speed = self.setpoint.get(what="speed")
            if self.detection_list.state == DetectionState.DANGER:
                self.setpoint.stop()
                log_detections = True
                logger.info("⛔️ danger detected, stopping.")
            elif self.detection_list.state == DetectionState.WARNING and speed > 0:
                self.setpoint.put(
                    what="speed",
                    value=speed // 2,
                )
                log_detections = True
                logger.info("⚠️ warning detected, lowering speed.")

            if env.BLUER_UGV_ULTRASONIC_SENSOR_LOG == 1:
                log_detections = False

            if log_detections:
                logger.info(
                    "{}: {}".format(
                        self.__class__.__name__,
                        ", ".join(self.detection_list.as_str(short=True)),
                    )
                )



from typing import List, Iterable, Iterator
from bluer_ugv.swallow.session.classical.ultrasonic_sensor.detection import (
    Detection,
    DetectionState,
)


class DetectionList:
    def __init__(
        self,
        list_of_detections: Iterable[Detection] | None = None,
    ):
        self._content: List[Detection] = (
            list(list_of_detections) if list_of_detections else []
        )

    def __iter__(self) -> Iterator[Detection]:
        return iter(self._content)

    def __len__(self) -> int:
        return len(self._content)

    def __getitem__(self, index: int) -> Detection:
        return self._content[index]

    def append(self, detection: Detection) -> None:
        self._content.append(detection)

    def as_str(
        self,
        short: bool = False,
    ) -> List[str]:
        return [detection.as_str(short=short) for detection in self._content]

    @property
    def state(self) -> DetectionState:
        if any(detection.state == DetectionState.DANGER for detection in self._content):
            return DetectionState.DANGER

        if any(
            detection.state == DetectionState.WARNING for detection in self._content
        ):
            return DetectionState.WARNING

        return DetectionState.CLEAR
